#include "Assym.hpp"
#include "Logging.hpp"
#include <fstream>

const char*  assym_cpp_version(){return "$Revision: 1.5 $";}

using std::ofstream;


void Assym::undoAssCompl(){
	clearProp();
	undoAss();
	LOG3("All ass undone");
}


void Assym::undoAss(){
	LOG2("Clear assigns of " << ass_stck.size() << " vars");
	for(uint32_t i=0;i < ass_stck.size();++i){
		sstate.vals.varClearValue(ass_stck[i].toVar());
	}
	ass_stck.clear();
	LOG2("Clear done ");
}

void Assym::clearProp(){
	prop_touched.clear();
	for(uint32_t i=prop_confl.size();i--;){
		sstate.cls_reg.cls_extr.setMrk(prop_confl[i],false);}
	prop_confl.clear();
	confl_strt = 0;
}


bool Assym::enqBcpBin(Literal tlit, const levl_t& reason, const levl_t& root){
	uint32_t bq			= ass_stck.size(); asrt(sstate.vals.litUkwn(tlit));
	enqueue(tlit,reason,root);asrt(sstate.vals.litTrue(tlit));
	Literal oth, flit;

	while(bq !=ass_stck.size()){
		tlit 			= ass_stck[bq++];
		flit 			= tlit.flipped();
		Iter<Literal> i	= sstate.wtch.getBinsOf(flit);
		while(i.getNext(oth)){
			if(sstate.vals.litFalse(oth)){ 	return false;}

			if(sstate.vals.litUkwn(oth)){		enqueue(oth,reason,root);
			}else{						vInfP(oth)->root |= root;}

			prop_touched.push_back(std::make_pair<Literal,ClsBin>(oth,ClsBin(flit)));
		}
	}
	return true;
}


//lits are not initialised. Literals are copied from orig ot to lits array
Assym::Task Assym::enqFlippedLits(Literal* lits, uint32_t& sz, levl_t& all_lvls, const ClsPointer& orig){
	levl_t levl = 1; sz=0; all_lvls=0; sanity(var_flgs.allCleared()); LOG3("enqFlippedLits initial size " << sz);
	//uint32_t max_mrk=0;

	SharedClause * cls = sstate.cinf.getCls(orig);

	FOR_LITS_STRT(cls,cl);

		//lit can be omitted
		if(sstate.vals.litFalse(cl)){sstate.stats->assym_shrnk_easy++; continue;}

		//clause is completely subsumed by binary (closure)
		if(sstate.vals.litTrue(cl)){ sstate.stats->assym_subsum_easy++;
			cls->setWaste(); 								return LEAVE;}

			//cls->setWaste("subsumed by bins",cl.toDIMACS()); return LEAVE;}

		//propagate flipped literal
		if(levl && !enqBcpBin(cl.flipped(),levl,levl)){
			LOG1("Unit [i] found in enqFlipped " << cl.toDIMACS());
			units_out.push_back(cl); 						return LEAVE;}

		asrt(sstate.vals.litFalse(cl) || levl==0); asrt(levl==0 || powOf2(vInfP(cl)->reasn));
		var_flgs	.setFlag(cl.toVar());
		lits[sz++] 	= cl;
		all_lvls 	|= levl;
		levl 		<<= 1;

	FOR_LITS_END;

	asrt(levl || sz >= sizeof(levl_t)*8);
	return sz > (sizeof(levl_t)*8) ? TAKE_MRKD : GO;
}



Assym::Task Assym::propLongs(const Literal* lits){
	Literal oth,flit,tlit; levl_t reason,roots; WtchEle ew; LOG3("propLongs " );

	for(uint32_t q=0; q!= ass_stck.size();++q){
		tlit				= ass_stck[q]; LOG2("Propagate lit " << tlit.toDIMACS());
		flit 				= tlit.flipped(); asrt(sstate.vals.litFalse(flit));
		WIter<WtchEle> i	= sstate.wtch.getWtchOfW(flit);

		while(i.getNext(ew)){
			if(sstate.cls_reg.cls_extr.isMrkd(ew.clp)){
				i.keep(ew);										continue;} //don't propagate on clause that are known to be false

			ClsWrp& cw		= sstate.cinf.get(ew.clp);
			oth				= cw.getOtherW(flit);

			if(sstate.vals.litTrue(oth)){
				if(var_flgs.isSet(oth.toVar())){ asrt(powOf2(vInfP(oth)->reasn));
					prop_touched.push_back(std::make_pair<Literal,ClsBin>(oth,ClsBin(ew.clp)));}
				i.keep(ew);										continue;}


			SharedClause* cls	= cw.getCls();
			ClsLit* beg			= cls->beg();
			ClsLit* iter		= cw.addOffset(beg);
			ClsLit* strt		= iter;

			do{
				Literal cand(iter->lit);

				if(!sstate.vals.litFalse(cand) && cand != oth){
					cw		.setWatchers(oth,cand);
					sstate.wtch	.addCls(cand,ew.clp,oth);
					cw		.keepOffset(beg,iter);
					goto NEXT_CLS;}

				iter = (iter->has_next) ? (iter+1) : beg;
			}while(iter != strt);

			//no new watcher found
			i.keep(ew);
			if(sstate.vals.litUkwn(oth)){
				collReasons(cls,oth,reason,roots);

				if(roots!=0){ addBins(oth,roots,lits);}
				asrt(!var_flgs.isSet(oth.toVar()));

				if(!enqBcpBin(oth,reason,roots)){
					LOG1("Unit found in propagation" << oth.flipped().toDIMACS());
					units_out.push_back(oth.flipped());
					while(i.getNext(ew)){i.keep(ew);}	sstate.wtch.finish(flit,i);
					return LEAVE;} asrt(sstate.vals.litTrue(oth));
			}else{
				LOG2("Conflict in propLong at cls " << ew.clp.idx());
				prop_confl.push_back(ew.clp);
				sstate.cls_reg.cls_extr.setMrk(ew.clp,true);
			}
			NEXT_CLS:;
		}
		sstate.wtch.finish(flit,i);
	}
	return GO;
}




//returns max32 if nothing changed
//or number of marked variables in orig clause that shall be kept
uint32_t Assym::shrinkyWinky(const ClsPointer& orig){
	asrt(units_out.empty()); //asrt(clses.asrtNoMrkdClses());

	LOG3("start shrinkyWinky for cls " << orig.idx() );
	levl_t all_levls; uint32_t mrkd;	sstate.stats->nbr_assym_chk++;

								LOG3("load extra infos" );
	ClsExtr& ec 				= sstate.cls_reg.cls_extr.get(orig); //asrt(ec.shallAssymChd()); asrt(!ec.isWaste());
	const uint32_t alloc_sz 	= ec.sz;
	ec.mrk						= true;
	Literal lits				[alloc_sz]; LOG3("allocate " << alloc_sz << " space " );

	Task task		= enqFlippedLits(lits,mrkd,all_levls,orig); asrt(task != NEWVERSION);

	uint32_t end	= mrkd; asrt(end<=alloc_sz);
	if(task!=GO){	goto FINISH_CLS;}

	do{
		task			= propLongs(lits); asrt(task != NEWVERSION);
		if(task!=GO){	goto FINISH_CLS;}

		task			= handleConflictingFor(orig,lits,end,mrkd,all_levls); asrt(end<=alloc_sz);

	}while(task==NEWVERSION)	; asrt(mrkd<=end);

	if(task==GO){
		task			= smpleChk(mrkd,lits,end);
	}

	FINISH_CLS:; asrt(mrkd<=end); asrt(task == LEAVE || task == TAKE_MRKD);
	undoAssCompl();

	if((task==TAKE_MRKD) && mrkd!=alloc_sz){ asrt(mrkd>0);
		LOG2("keep " << mrkd << " lits instead of " << alloc_sz);
		sstate.stats->assym_lits_rem += (alloc_sz - mrkd);

		mrkd = 0;
		for(uint32_t i=0;i!=end;++i){
			if(var_flgs.isSet(lits[i].toVar())){
				var_flgs		.clear(lits[i].toVar());
				lits[mrkd++] 	= lits[i];}
		}
		sanity(var_flgs.allCleared());
		LOG2("filtered lits to  " << mrkd );

		SharedClause* ocls 		= sstate.cinf.getCls(orig);
		if(mrkd==1){
			LOG1("Unit in shrinkyWinky: " << lits[0].toDIMACS());
			units_out.push_back(lits[0]);
		}else if(mrkd==2){
			LOG2("Binary in shrinkyWinky: " << lits[0].toDIMACS() << " , " << lits[1].toDIMACS());
			sstate.addBinAndSend(lits[0],lits[1]);
			ocls->setWaste();
			//ocls->setWaste("Subsumed by bin orig was ",alloc_sz);
		}else{
			SharedClause* ncls	= SharedClause::construct(sstate.s_info->all_users_bits,
									lits,lits+mrkd,ocls->getOrigLBD(),ocls->isLearnt());
			sstate.linkCls(ncls,mrkd,lits[0],lits[1],ocls->getUsers());
			ocls->setWaste();
		}
	}else{
		LOG2("quit " << mrkd  );
		for(mrkd=0;mrkd!=end;++mrkd){var_flgs.carelessClear(lits[mrkd].toVar());}
	}

	sstate.cls_reg.cls_extr.get(orig).mrk = false; //asrt(clses.asrtNoMrkdClses());


	LOG3("finish shrinkyWinky");
	return max32; //no direct change but maybe there are unit clauses or subsumed clauses
}



void Assym::collReasons(SharedClause* cls, const Literal& except_w, levl_t& reason, levl_t& root){
	reason = root = 0;

	FOR_LITS_STRT(cls,lit);
		if(lit == except_w){continue;}

		VInf* vi	= vInfP(lit);
		reason		|= vi->reasn; asrt(!var_flgs.isSet(lit.toVar()) || powOf2(vi->reasn));
		root		&= vi->root;

	FOR_LITS_END;
}


uint32_t Assym::collRsnsCountMrkd(SharedClause* cls, levl_t& reason, levl_t& root){
	uint32_t m = reason=root=0; LOG2("Check cls at " << (uint64_t)cls);

	FOR_LITS_STRT(cls,lit);

		LOG2("Check lit " << lit.toDIMACS());
		Variable v 	= lit.toVar(); asrt(sstate.vals.varKnwn(v));
		VInf* vi	= vInfP(v); asrt(!var_flgs.isSet(v) || powOf2(vi->reasn));
		reason		|= vi->reasn;
		root		&= vi->root;
		m			+= var_flgs.isSet(v);

	FOR_LITS_END;

	return m;
}

void Assym::foundSubsumd(const ClsPointer& subsumed, const ClsPointer& by){
	if(sstate.cinf.getCls(subsumed)->isLearnt()) return;

	ClsExtr& sub   = sstate.cls_reg.cls_extr.get(by);
	if(sstate.cinf.getCls(by)->isLearnt()){
		sstate.ensureClsFix(by,sub.sz); LOG2("Change to fix cls " << by.idx());
		asrt(sstate.cinf.getCls(by)->isFix());
	}
}

Assym::Task 	Assym::handleConflictingFor(const ClsPointer& orig,Literal* lits, uint32_t& end, uint32_t& mrkd, levl_t& all_levs){
	levl_t reason,roots; LOG3("handleConflicting for " <<orig.idx() );
	levl_t min_reason = all_levs;
	uint8_t c_min_r	= count_ones.get64(min_reason); asrt (c_min_r<=end); asrt(confl_strt <= prop_confl.size());

	while(confl_strt<prop_confl.size()){
		ClsPointer& clp		= prop_confl[confl_strt++]; asrt(orig.idx() != clp.idx());
		SharedClause* cls	= sstate.cinf.getCls(clp); LOG2("Collect reasons of false " << clp.idx());
		uint32_t m			= collRsnsCountMrkd(cls,reason,roots);

		if(roots){//conflict depends on at least one unit cls
			for(uint32_t i=0;roots;++i,roots>>=1){
				if(roots&1){ LOG1("Unit found by roots " << lits[1].toDIMACS());
					units_out.push_back(lits[i]); }
			}
			return LEAVE;
		}

		ClsExtr& ec			= sstate.cls_reg.cls_extr.get(clp); asrt(ec.mrk);
		if(m==end){ //current cls is subsumed by orig
			LOG2("Cls " << clp.idx() << " is subsumed by original " << orig.idx());
			foundSubsumd(clp,orig);

		}else if(m == ec.sz){//all lits marked in cls -> cls subsumes orig
			LOG2("Original " << clp.idx() << " is subsumed by cls " << orig.idx());
			foundSubsumd(orig,clp);
			return LEAVE;
		}

		if(reason!=all_levs){ asrt((reason & ~all_levs) == 0); //conflict depends on subset of lits

			uint8_t invlvd	= count_ones.get64(reason);
			if(invlvd < c_min_r){c_min_r=invlvd; min_reason=reason;}
		}
	}

	if(min_reason == all_levs){ return GO;} //all conflicts depend on all lits

	//propagate the flipped lits of the new subclause again
	undoAss();
	prop_touched.clear(); asrt(end <= sizeof(levl_t)*8);
	mrkd = 0; levl_t levl = 1; all_levs = 0;
	LOG3("Propagate lits in new subclause again");

	for(uint32_t l=0;l!=end;++l,min_reason>>=1){
		Literal cl = lits[l]; var_flgs.clear(cl.toVar());

		//lit can be omitted
		if(sstate.vals.litFalse(cl)){sstate.stats->assym_shrnk_easy++;	continue;}

		//clause is completely subsumed by binary (closure)
		if(sstate.vals.litTrue(cl)){ sstate.stats->assym_subsum_easy++;
			sstate.cinf.getCls(orig)->setWaste(); 					return LEAVE;}

			//sstate.cinf.getCls(orig)->setWaste("Assym handle conflict shrink ",cl.toDIMACS()); return LEAVE;}


		asrt(sstate.vals.litUkwn(cl));
		if(min_reason&1){
			var_flgs.setFlag(cl.toVar()); //should be already set
			lits[mrkd++]=cl;

			if(levl && !enqBcpBin(cl.flipped(),levl,levl)){
				LOG1("Unit [i] found in re enqFlipped " << cl.toDIMACS());
				units_out.push_back(cl); 	return LEAVE;}

			all_levs 	|= levl;
			levl 		<<= 1;
		}
	} asrt(min_reason == 0);

	end = mrkd; LOG2("New version available with size " << mrkd);
	return NEWVERSION;
}


Assym::Task 	Assym::smpleChk(uint32_t& mrkd_c,const Literal* lits,const uint32_t& end){
	mrkd_c = end;
	if(mrkd_c <= 2) 	return TAKE_MRKD;

	levl_t rem_lev 	= 0;levl_t alt_reas,alt_roots; bool change;
	uint32_t nsz = prop_touched.size();

	do{
		change 				= false;
		const uint32_t sz 	= nsz;
		nsz					= 0;


		for(uint32_t t=0;t!= sz;++t){

			Literal lu		= prop_touched[t].first; asrt(sstate.vals.litTrue(lu));
			ClsBin&	clw		= prop_touched[t].second;

			if(!getAltReason(lu,clw,alt_reas,alt_roots)) continue;

			VInf* ai		= vInfP(lu); LOG2("Reason of " << lu.toDIMACS() << " is " << ai->reasn << " alt is " << alt_reas << " removed are " << rem_lev);
			addBins(lu,ai,alt_roots,lits);asrt((rem_lev & alt_reas) ==0);

			//var(lu) is not a variable of the remaining clause
			if(!var_flgs.isSet(lu.toVar())){			continue;}
			asrt(powOf2(ai->reasn)); 					//just itself is reason

			//check if there is an alternative valid binary reason
			levl_t binr = ai->root & ~(rem_lev | ai->reasn);
			if(binr){
				//next check will match since alt_reas is independent of lu
				for(alt_reas=1;(binr&alt_reas)==0;alt_reas<<=1){} asrt((ai->reasn & alt_reas) == 0);}

			if((ai->reasn & alt_reas) == 0){ //others are independent of lu
				var_flgs.clear(lu.toVar()); 	asrt((rem_lev & ai->reasn) ==0);
				rem_lev |= ai->reasn; 			asrt((rem_lev & alt_reas) ==0);
				change	= true;
				LOG2("Change reasons of " << lu.toDIMACS() << " to " << alt_reas << " rem are " << rem_lev	);

				updateAllReasons(lits,end,ai->reasn,alt_reas);
				asrt((ai->reasn & rem_lev) == 0); //sstate.stats->assym_shrnk_ana++;

				if(--mrkd_c <= 2){ return TAKE_MRKD;}
			}else{
				prop_touched[nsz++] = prop_touched[t];
			}
		}
	}while(change);// && glob.sets.assym_loop);

	return TAKE_MRKD;
}



bool	Assym::getAltReason(const Literal& except, const ClsBin& clw,levl_t& reas_out, levl_t& root_out){ asrt(sstate.vals.litTrue(except));
	if(clw.isBin()){ asrt(sstate.vals.litFalse(clw.getLit()));
		VInf* ai 	= vInfP(clw.getLit());
		reas_out	= ai->reasn;
		root_out	= ai->root;	LOG3("Set reason by bin " << clw.getLit().toDIMACS());
		return true;
	}

	reas_out=root_out=0; LOG3("Set reasons first to " << reas_out << " and " <<root_out);
	SharedClause* cls = sstate.cinf.getCls(clw.getClp());
	FOR_LITS_STRT(cls,cnd);
		if(cnd==except){continue;}
		if(!sstate.vals.litFalse(cnd)){ LOG3("Cls not unit");return false;}
		VInf* aj	= vInfP(cnd);
		reas_out	|= aj->reasn;
		root_out	&= aj->root;	LOG3("Lit " << cnd.toDIMACS() << " with reason " << aj->reasn << " change to " << reas_out);
	FOR_LITS_END;

	return true;
}

void 	Assym::updateAllReasons(const Literal* lits, const uint32_t& end, const levl_t old, const levl_t repl){
	asrt(powOf2(old)); asrt((old&repl)==0);
	LOG2("update all reasons remove " << old << " repl by new " << repl);

	levl_t newl;
	for(uint32_t j=0;j!=end;++j){ asrt(sstate.vals.litFalse(lits[j]));
		VInf* ti		= vInfP(lits[j]);
		if(ti->reasn & old) {
			newl		= (ti->reasn & ~old) | repl;
			ti->reasn	= newl; LOG3("set reason of " << lits[j].toDIMACS() << " to " << ti->reasn);
		}asrt((ti->reasn & old) == 0);
	}

	for(uint32_t i=0;i!=ass_stck.size();++i){
		VInf* ai	= vInfP(ass_stck[i]);
		if(ai->reasn & old) {
			newl		= (ai->reasn & ~old) | repl;
			ai->reasn	= newl; LOG3("set reason of " << ass_stck[i].toDIMACS() << " to " << ai->reasn);
		}asrt((ai->reasn & old) == 0);
	}
}

#ifdef ASSYM_HYPER_BIN
void Assym::addBins(const Literal& lit, levl_t roots, const Literal* lits){
	asrt(roots); asrt(!sstate.vals.litFalse(lit));
	//const bool add = act_hp.getValue(lit.toVar()) >= hbin_over;
	//if(act_hp.getValue(lit.toVar()) < hbin_over) return;

	for(uint32_t i=0;roots;++i,roots>>=1){
		if(roots&1){
			LOG2("Binary in Assym by roots : " << lits[i].toDIMACS());
			sstate.addBinAndSend(lits[i],lit);
			sstate.stats->nbr_hyper_bin_res++;}
	}
}

void Assym::addBins(const Literal& lit, VInf* ai, levl_t roots, const Literal* lits){
	roots 	&= ~(ai->root); //are there new binaries
	if(roots){
		ai->root |= roots; //store new binary implications
		addBins(lit,roots,lits);
	}
}
#else
void Assym::addBins(const Literal& lit, levl_t roots, const Literal* lits){}
void Assym::addBins(const Literal& lit, VInf* ai, levl_t roots, const Literal* lits){}
#endif

/*void Assym::justHpyerBin(const vector<Literal>& units_in, vector<Literal>& units_out,  const Literal& root){
	for(;units_updated < units_in.size();++units_updated){
			sstate.vals.setValue(units_in[units_updated]);}
	asrt(units_out.empty()); asrt(sstate.vals.litUkwn(root)); asrt(prop_confl.empty());

	if(!enqBcpBin(root,1,1)){
		units_out.push_back(root.flipped());
	}else{
		vector<Literal> lits(1,root.flipped());
		Literal res 	= propOfLits(ClsP(),lits);
		if(Literal::nrmlLit(res)){
			units_out.push_back(res.flipped());}
		if(!prop_confl.empty()){
			units_out.push_back(root.flipped());}

	}

	undoAssigns();
}*/





