/*
 * SatTypes.hpp
 *
 *  Created on: 13.08.2009
 *      Author: stephan
 */

#ifndef SATTYPES_HPP_
#define SATTYPES_HPP_
#define sattypes_hpp_version      "$Revision: 1.2 $"

#include "GlobalSettings.hpp"
#include <vector>


typedef LazyFraction<7,uint16_t,uint16_t> cls_act_t;


class Literal;


/**Class to represent boolean values and a third state 'Unknown'*/
class TrBool{
	uint8_t val;

//public members
public:
	static const uint8_t   FALSE_TRBOOL_VAL		= 0;
	static const uint8_t   TRUE_TRBOOL_VAL		= 1;
	static const uint8_t   TRBOOL_FLIP_MASK		= 1;
	static const uint8_t   UNKNOWN_TRBOOL_VAL	= 2;

	inline			TrBool();

	/**Constructs an object with the specified value.
	* @param i: a value from {0,1,2}*/
	inline explicit	TrBool(uint8_t i);

	inline			TrBool(const TrBool& b);

	//inline uint8_t	toInt()							const;
	inline bool		isTrue()						const;
	inline bool		isFalse()						const;
	inline bool		isUnknown()						const;
	inline bool		isKnown()						const;
	inline bool		operator==(const TrBool& o)		const;
	inline bool		operator!=(const TrBool& o)		const;
};


/**Class to represent variables. All additional information about a variable are stored outside: in an instance of
 * ConstraintDB*/
class Variable{

	uint32_t ID;

public:
	inline explicit Variable();
	inline explicit	Variable(const uint32_t i);

	inline uint32_t	idx()							const;
	inline Literal	toLit(const bool b)				const;
	inline Literal	toPosLit()						const;
	inline Literal	toNegLit()						const;
	inline uint32_t	toDIMACS()						const;
	inline bool		operator<(const Variable& v)	const;
	inline bool		operator!=(const Variable& v)   const;
	inline bool		operator==(const Variable& v)   const;

	friend class Literal;
};


/**Class to represent literals. All aditional information are stored elsewhere: in an instance of class ConstraintDB*/
class Literal{

	uint32_t ID;

public:
	/**Return a literal object from the given DIMACS number*/
	inline static Literal DIMACS2Literal(int32_t d);

	/**Construct an invalid literal*/
	inline				Literal();

	/**Construct a literal with the specified id*/
	inline explicit		Literal(const uint32_t i);

	inline uint32_t		idx()			const;
	inline Literal		flipped()		const;
	inline Variable		toVar()			const;
	inline bool			isPos()			const;
	inline bool			isNeg()			const;
	inline TrBool		pol()			const;
	inline TrBool		flpdPolarity()	const;
	inline int32_t		toDIMACS()		const;

	/**Compare literals by their ids*/
	inline bool			operator<(const Literal& l) const;
	inline bool			operator>(const Literal& l) const;
	inline bool			operator!=(const Literal& l) const;
	inline bool			operator==(const Literal& l)const;

	friend class Variable;
};


typedef vector<Literal>::iterator liter_t;
typedef vector<Variable>::iterator viter_t;



//implementation

const Literal   INVALID_LIT			(max32);
const Literal	FALSE_LIT			(0);
const Literal	TRUE_LIT			(FALSE_LIT.flipped());
const Variable  INVALID_VAR			(max32);
const TrBool	TRUE_TrBool			(TrBool::TRUE_TRBOOL_VAL);
const TrBool	FALSE_TrBool		(TrBool::FALSE_TRBOOL_VAL);
const TrBool	UNKNOWN_TrBool		(TrBool::UNKNOWN_TRBOOL_VAL);



//TBool Implementation
			TrBool::TrBool()						:val(UNKNOWN_TRBOOL_VAL){ }
			TrBool::TrBool(const TrBool& b)			:val(b.val)			{ }
			TrBool::TrBool(uint8_t i)				:val(i)	{
							asrt(i == TRUE_TRBOOL_VAL || i == FALSE_TRBOOL_VAL || i == UNKNOWN_TRBOOL_VAL);}

//uint8_t		TrBool::toInt()				const	{return val;}
bool		TrBool::isTrue()			const	{return val == TRUE_TRBOOL_VAL;}
bool		TrBool::isFalse()			const	{return val == FALSE_TRBOOL_VAL;}
bool		TrBool::isUnknown()			const	{return val == UNKNOWN_TRBOOL_VAL;}
bool		TrBool::isKnown()			const	{return val != UNKNOWN_TRBOOL_VAL;}
bool		TrBool::operator==(const TrBool& o) const   {return val == o.val;}
bool		TrBool::operator!=(const TrBool& o) const   {return val != o.val;}



//Variable Implementation
			Variable::Variable() 							:ID(~zero32){ }
			Variable::Variable(const uint32_t i) 			:ID(i){ }
Literal		Variable::toLit(const bool b)			const   {return Literal((ID << 1) + b);}
Literal		Variable::toPosLit()					const   {return Literal((ID << 1) + 1);}
Literal		Variable::toNegLit()					const   {return Literal(ID << 1);}
uint32_t	Variable::idx()							const   {return ID;}
uint32_t	Variable::toDIMACS()					const   {return ID;}
bool		Variable::operator<(const Variable& v)  const   {return ID < v.ID;}
bool		Variable::operator!=(const Variable& v) const   {return ID != v.ID;}
bool		Variable::operator==(const Variable& v) const   {return ID == v.ID;}

//Literal Implementation
			Literal::Literal()						:ID(~zero32){ } //invalid lit
			Literal::Literal(const uint32_t i)   	:ID(i)	{ }
uint32_t	Literal::idx()							const   {return ID;}
Literal		Literal::flipped()						const   {return Literal(ID ^ 1);}
Variable	Literal::toVar()						const   {return Variable(ID >> 1);}
bool		Literal::isPos()						const   {return ID & 1;}
bool		Literal::isNeg()						const   {return !(ID & 1);}
int32_t		Literal::toDIMACS()						const   {return ID&1 ? ID>>1 : -(ID>>1);}
TrBool		Literal::pol()							const   {return TrBool(ID & 1);}
TrBool		Literal::flpdPolarity()					const   {return TrBool(!(ID&1));}
bool		Literal::operator==(const Literal& l)   const   {return ID == l.ID; }
bool		Literal::operator<(const Literal& l)	const   {return ID < l.ID;}
bool		Literal::operator>(const Literal& l)	const   {return ID > l.ID;}
bool		Literal::operator!=(const Literal& l)   const   {return ID != l.ID;}

Literal	Literal::DIMACS2Literal(int32_t d){
	asrt(d != 0 );
	int32_t m(d>>31);
	return  Literal(((d^m)<<1)+1-m);
}


struct EmptyVisitor{
	const bool operator()(const Literal& l) const{return true;}
	const bool operator()(const Literal& l,const Literal& b) const{return true;}
};





#endif /* SATTYPES_HPP_ */
