/*
 * SharedBasic.hpp
 *
 *  Created on: Nov 27, 2009
 *      Author: kottlers
 */

#ifndef SHAREDBASIC_HPP_
#define SHAREDBASIC_HPP_
#define sharedbasic_hpp_version "$Revision: 1.3 $"

#include "GlobalSettings.hpp"
#include "Logging.hpp"


inline bool exchange8(volatile uint8_t* varp,const uint8_t assume, const uint8_t new_val){
	return __sync_bool_compare_and_swap(varp,assume,new_val);}


inline bool exchange16(volatile uint16_t* varp,const uint16_t assume, const uint16_t new_val){
	return __sync_bool_compare_and_swap(varp,assume,new_val);}
	//return *varp!=assume?false : (*varp=new_val,true);}


inline bool exchange32(volatile uint32_t* varp,const uint32_t assume, const uint32_t new_val){
	return  __sync_bool_compare_and_swap(varp,assume,new_val);
}
	//return *varp!=assume?false : (*varp=new_val,true);}

/*inline bool exchange64(volatile uint64_t* varp,const uint64_t assume, const uint64_t new_val){
	return __sync_bool_compare_and_swap(varp,assume,new_val);}*/


#define MAX_USE_THREADS 16

struct UserID{
	uint16_t id;
	explicit UserID(const uint32_t thread_id):id(thread_id){
			asrt(thread_id <= MAX_USE_THREADS);}
};

inline bool operator==(const UserID& a, const UserID& b){return a.id==b.id;}
inline bool operator!=(const UserID& a, const UserID& b){return a.id!=b.id;}



struct UserBit{
	const uint16_t bit_mask;
	explicit UserBit(const UserID& uid):bit_mask(1<<uid.id){}
};

struct UserMask{
	uint16_t mask;

	explicit UserMask(const UserBit& one):mask(one.bit_mask){};
	explicit UserMask(const uint16_t& users):mask(users){};

	bool hasAnyUser()const					{return mask != 0;}
	bool hasUser(const UserBit& b) const 	{return mask & b.bit_mask;}
	bool isOnlyUser(const UserBit& b)const	{return (mask & ~b.bit_mask) == 0;}


	void addUser(const UserBit& b) {
		asrt(!(mask & b.bit_mask));

		uint16_t	curr_users 		= mask;
		while(!exchange16(&mask,curr_users,curr_users | b.bit_mask)){
			#pragma omp flush (mask)
			LOG3("Try to add user");
			curr_users 	= mask;}
	}

	bool removeUser(const UserBit& ub){
		asrt(mask & ub.bit_mask);

		const uint16_t bit_id_inv = ~ub.bit_mask;
		uint16_t	curr_users 		= mask;
		uint16_t 	new_users		= curr_users & bit_id_inv;

		while(!exchange16(&mask,curr_users,new_users)){
			#pragma omp flush (mask)
			LOG3("Try to remove user");
			curr_users 	= mask;
			new_users 	= curr_users & bit_id_inv;}

		return new_users == 0;
	}
};

//inline bool operator==(const UserMask& a, const UserMask& b) {return a.mask == b.mask;}




template<class T>
class UpdateChain{
	typedef T subclass;
protected:
	static const uint32_t	UPDATE_BIT 		= 1 << 16;
	static const uint32_t	USERS_MASK 		= UPDATE_BIT - 1;
	static const uint32_t	MAX_EXTRA		= (1 << 15)-1;
	static const uint32_t	UU_MASK 		= (UPDATE_BIT | USERS_MASK);
	//static const uint32_t	CREATOR_MASK	= ((1 << 20) - 1) ^ USERS_MASK;

	T*						update;
	volatile uint32_t 		ext15_udt1_use16;

	inline uint32_t clearBit(const uint32_t& bit);

	static bool hasUpdate(const uint32_t u)	{return u & UPDATE_BIT;}
	static bool hasAnyUser(const uint32_t u){return u & USERS_MASK;}
protected:
	bool hasUser(const UserBit& ub) const{	return ext15_udt1_use16 & ub.bit_mask;}

	bool removeUser(const UserBit& ub){
		asrt(hasUser(ub));
		return !hasAnyUser(clearBit((uint32_t)ub.bit_mask));}

	static void forgetThis(T* t, const UserBit&  ub){
		asrt(t!= NULL);
		if(t->removeUser(ub)){T::destruct(t);}}
public:
	bool hasUpdate() const		{return hasUpdate(ext15_udt1_use16);}

protected:
	T*	getUpdateLink() const	{asrt(hasUpdate()); asrt(update != NULL);return update;}
	uint32_t getExtra() const	{return ext15_udt1_use16 >> 17;}
	bool	setExtraIf(const uint32_t assume, const uint32_t set){
			uint32_t a = (ext15_udt1_use16&UU_MASK)|(assume<<17); 	asrt(assume<=MAX_EXTRA);
			uint32_t s = (ext15_udt1_use16&UU_MASK)|(set<<17);		asrt(set <=MAX_EXTRA);
			return exchange32(&ext15_udt1_use16,a,s);}

	UpdateChain(const UserMask& users,const uint32_t extra = 0)
	:update(NULL),ext15_udt1_use16((uint32_t)users.mask | (extra<<17)){
		asrt(getExtra()==extra);asrt(hasAnyUser(users.mask));}

	UserMask getAllUsers(){
		asrt(!hasUpdate()); return UserMask(ext15_udt1_use16&(uint32_t)max16);}

	inline static void forgetAll(T* t, const UserBit&  ub);
	inline static T* loadNextUpdate(T* t,const UserBit&  u_bit);
	inline static T* loadAvailUpdate(T* t,const UserBit&  u_bit);

	//only for one thread
	inline void linkUpdate(T* updt);
};

template<class T>
uint32_t UpdateChain<T>::clearBit(const uint32_t& bit){
	asrt(ext15_udt1_use16 & bit);

	const uint32_t bit_id_inv = ~bit;
	uint32_t	curr_mask 		= ext15_udt1_use16;
	uint32_t 	new_mask		= curr_mask & bit_id_inv;

	while(!exchange32(&ext15_udt1_use16,curr_mask,new_mask)){
		#pragma omp flush (ext15_udt1_use16)
		//LOG_DETAIL("Try to unlink Bin Edges");
		curr_mask 		= ext15_udt1_use16;
		new_mask		= curr_mask & bit_id_inv;}

	return new_mask;
}

template<class T>
void UpdateChain<T>::linkUpdate(T* updt){asrt(update == NULL); asrt(!hasUpdate());
	uint32_t	mask;
	update		= updt;

	do{
		#pragma omp flush (ext15_udt1_use16)
		mask 					= ext15_udt1_use16;
		updt->ext15_udt1_use16	= mask; asrt(hasAnyUser(mask));
		asrt(!hasUpdate());
	}while(!exchange32(&ext15_udt1_use16,mask,mask | UPDATE_BIT));
}

template<class T>
void UpdateChain<T>::forgetAll(T* t, const UserBit&  ub){
	asrt(t->hasUser(ub));
	do{
		uint32_t new_mask = t->clearBit((uint32_t)ub.bit_mask);
		T* next = hasUpdate(new_mask) ? t->getUpdateLink() : NULL;

		if(!hasAnyUser(new_mask)){T::destruct(t);}
		t = next;
	}while(t != NULL);
}

template<class T>
T* UpdateChain<T>::loadNextUpdate(T* t,const UserBit&  u_bit){
	asrt(t != NULL); asrt(t->hasUser(u_bit));
	if(t->hasUpdate()){
		T* next = t->getUpdateLink(); asrt(next != NULL);
		forgetThis(t,u_bit);
		asrt(next->ext15_udt1_use16 & u_bit.bit_mask);
		return next;}
	return t;
}

template<class T>
T* UpdateChain<T>::loadAvailUpdate(T* t,const UserBit&  u_bit){
	asrt(t != NULL); asrt(t->hasUser(u_bit));
	while(t->hasUpdate()){
		T* next = t->getUpdateLink(); asrt(next != NULL);
		forgetThis(t,u_bit);
		t = next; asrt(t->ext15_udt1_use16 & u_bit.bit_mask);}
	return t;
}



template<class P>
class FlagPtr{
	static const uint64_t FLAG_BIT			= 1;
	static const uint64_t VAL_BIT			= 2;
	static const uint64_t PTR_MASK			= ~zero64 - 3;

	uint64_t	data;
protected:
	explicit 	FlagPtr(const uint64_t& d):data(d){}
	uint64_t	toInt() const	{return data;}
public:
	bool		isVal() const	{return data & VAL_BIT;}
	bool		isPtr() const	{return !(data & VAL_BIT);}
	P*			ptr()	const	{asrt(isPtr()); return (P*)(data&PTR_MASK);}
	uint64_t	val()	const	{asrt(isVal()); return data >> 2;}

	bool		isFlagged()const{return data & FLAG_BIT;}
	void		setFlag()		{data|= FLAG_BIT;}
	void		unsetFlag()		{data &= ~FLAG_BIT;}


	FlagPtr(const P* t, const bool flg)
		:data(((uint64_t)t)|((uint64_t)flg)){
		asrt(isPtr());asrt(isFlagged()==flg);asrt(ptr()==t);}
	FlagPtr(const uint64_t& v, const bool flg)
		:data((((uint64_t)v)<<2)|VAL_BIT|((uint64_t)flg)){
		asrt(isVal());asrt(v==val());asrt(isFlagged()==flg);}

	void		changeToPtr(const P* t)
				{data	= ((uint64_t)t) | (data&FLAG_BIT);asrt(isPtr());asrt(ptr()==t);}
	void		changeToPtrDelFlag(const P* t)
				{data	= (uint64_t)t; asrt(isPtr());asrt(ptr()==t);}
	void		changeToPtrSetFlag(const P* t)
				{data	= ((uint64_t)t) | FLAG_BIT; asrt(isPtr());asrt(ptr()==t);}
	void		changeToVal(const uint64_t& i){
				data	= (((uint64_t)i)<<2) | VAL_BIT | (data&FLAG_BIT);asrt(isVal());asrt(val()==i);}
	void		changeToValDelFlag(const uint64_t& i){
				data	= (((uint64_t)i)<<2) | VAL_BIT;asrt(isVal());asrt(val()==i);}
	void		changeToValSetFlag(const uint64_t& i){
				data	= (((uint64_t)i)<<2) | VAL_BIT | FLAG_BIT;asrt(isVal());asrt(val()==i);}

	FlagPtr(const FlagPtr& p):data(p.data){}
	FlagPtr& operator=(const FlagPtr& p){data = p.data;return *this;}
};


template<class P1, class P2>
class WrapPtr{
	static const uint64_t PTR1_BIT			= 1;
	static const uint64_t PTR2_BIT			= 2;
	static const uint64_t TYPE_MASK			= 3;
	static const uint64_t PTR_MASK			= ~zero64 - 3;

	uint64_t	data;
protected:
	explicit 	WrapPtr(const uint64_t& d):data(d){}
	uint64_t	toInt() const	{return data;}
public:
	bool		isAnyPtr()const	{return data & TYPE_MASK;}
	bool		isVal() const	{return !isAnyPtr();}
	bool		isPtr1()const	{return data & PTR1_BIT;}
	bool		isPtr2()const	{return data & PTR2_BIT;}
	P1*			ptr1()	const	{asrt(isPtr1()); return (P1*)(data&PTR_MASK);}
	P2*			ptr2()	const	{asrt(isPtr2()); return (P2*)(data&PTR_MASK);}
	uint64_t	val()	const	{asrt(isVal()); return data >> 2;}

	static WrapPtr construct(const P1* t){return WrapPtr(((uint64_t)t)|PTR1_BIT);}
	static WrapPtr construct(const P2* t){return WrapPtr(((uint64_t)t)|PTR2_BIT);}
	static WrapPtr construct(const uint64_t& v){return WrapPtr(v<<2);}

	void		changeToPtr(const P1* t)
				{data	= ((uint64_t)t) | PTR1_BIT;asrt(isPtr1());asrt(ptr1()==t);}
	void		changeToPtr(const P2* t)
				{data	= ((uint64_t)t) | PTR2_BIT;asrt(isPtr2());asrt(ptr2()==t);}
	void		changeToVal(const uint64_t& i){
				data	= ((uint64_t)i)<<2;asrt(isVal());asrt(val()==i);}

	WrapPtr(const WrapPtr& p):data(p.data){}
	WrapPtr& operator=(const WrapPtr& p){data = p.data;return *this;}
};






#endif /* SHAREDBASIC_HPP_ */
