
#ifndef CDCLCHIEF_HPP_
#define CDCLCHIEF_HPP_
#define solverleader_hpp_version      "$Revision: 1.29 $"

#include "GlobalSettings.hpp"
#include "SolverTrooper.hpp"
#include "Satistics.hpp"
#include "SCC.hpp"
#include "Assym.hpp"
//#include "Analysis.hpp"
//#include "Subsumption.hpp"

#define LEADERTEMP template<class BCPT>
#define LEADERCLASS SolverLeader<BCPT>

const uint32_t MAX_LITS		= 150;
const uint32_t MAX_FACT		= 12;
const uint32_t MAX_PRODUCT 	= MAX_FACT * MAX_FACT;



LEADERTEMP
struct SolverLeader:public BCPT{
	typedef  BCPT						super;
	volatile bool 						print_cons;


protected:
	ConflictsAndGC 						conflicts;
	Restarts							restart;
	volatile uint32_t					timeout;

	vector<Literal>						unpropd_units;
	BitVector<Variable>					var_flgs;

	Assym								assym;

	uint32_t 							simp_last_units;
	uint32_t							simp_scc_on_bins;
	//uint32_t							simp_last_max_lbd;
	//uint64_t							simp_next_earliest;
	uint64_t							simp_next_latest;
	uint32_t							occ_list_units_cleared;
	//Analysis							analysis;

	//clear clauses
	inline TrBool 	simplify();
	inline void		sendUnitsAt0();
	inline bool		propAt0(const bool send);
	inline bool		propAt0(vector<Literal>& lits);
	inline void		cleanClause(SharedClause* cls);

	inline TrBool	search();

	inline void		paraInit();

	inline SolverLeader(const GlobalSettings* sets, Satistics* sta, typename super::decision_t* dec, LeaderSiblingInfo* sinf,
			const Restarts& restart_init,const ConflictsAndGC& confl,const bool glue,
			const BeyondUP beyup,const HyperBinRes hbres,const OnTheFlySubsum otfsub);
	~SolverLeader(){}
public:
	static inline SolverLeader* create(const GlobalSettings* sets,const uint32_t grp_sz,const uint32_t t_id,ElimClses& eli,
			typename BCPT::decision_t::act_t max,typename BCPT::decision_t::act_t decay, ImportCriteria imprt,
			uint32_t rand_freq_permil,PolarityMode m,const Restarts& restart_init,const ConflictsAndGC& confl,
			const bool glue,  const BeyondUP beyup,const HyperBinRes hbres,const OnTheFlySubsum otfsub);
	static SolverLeader* destruct(SolverLeader* s){
			Satistics* st=s->super::sstate.stats;	LeaderSiblingInfo* sinf=s->super::sstate.s_info;
			typename super::decision_t* dec	=s->super::decision;
			delete s;  delete dec; delete sinf; delete st; return NULL;}


	inline TrBool 		isSat();
	inline void			printModel()					{super::printModel();}
	inline void			getModel(vector<bool>& var_vals){super::getModel(var_vals);}
	inline Literal 		validateCls(vector<Literal>& literals);
	inline bool			addValidUnsharedClause(const Literal& l0, const Literal& l1);
	inline void			addValidSharedClause(SharedClause* cls, const uint32_t sz);


	void				reserveForVariables(uint32_t s)	{
						super::reserveForVariables(s);
						var_flgs.ensureSize(s,false);}

	inline void			stopASAP();



protected:
	//inline uint32_t 	checkBinBWSubsumed(const SharedClause& cls,Literal cls_out[]);
	//inline bool		 	checkBinBWSubsumed(TempCls& cls);

	//inline uint32_t 	checkSubsumption(Literal lits[], const uint32_t& l_sz,SharedClause* ignore, const bool ensure_cls);
	//inline bool			checkSubsumption(SharedClause* cls);
	//inline uint32_t 	checkSubsumed(const Literal& lit, const uint32_t& lits_above, SharedClause*& dont);

	inline bool			enqueue(Literal l);
	//inline bool 		deriveShorts(SharedClause& cls);
	//inline bool			failedProp();
	//inline bool 		failedProp(const Literal& lit);

	/*inline TrBool		checkAutarky(const Literal lits[], const uint32_t sz);
	inline bool			checkAutarkies();*/
	inline bool			sccReplace();
	inline bool			subsumChks();
	inline bool 		subsumChkBins(uint32_t& cnt);
	inline Literal 		chkSuperToBins(const Literal& sub_bin, const SharedClause* super,const bool allow_flpd);
	inline bool 		collectBinSubsumed(const Literal& lit,vector<OccCls>& occs);
	inline bool			removeLit(SharedClause* cls, uint32_t sz, const Literal& rem);

	inline void			clearFlgOf(const SharedClause* cls);
	inline bool 		collectSubsumed(const ClsPointer& subp);
	inline Literal 		chkSuperOrEqToMarked(SharedClause* sup , uint32_t sub_mrkd);

	inline bool 		mrkFirstFactor(const ClsBin& cb,const Literal& skip, vector<Literal>& out);
	inline bool 		multiplyWithMrkd(const ClsBin& cb,vector<Literal>& out);
	inline void 		elimClses(ElimClses& eli, const Literal& lit, vector<ClsBin>& fakt);
	inline TrBool 		elimVarIf(const Variable& var,vector<Message>& eli_mess);

	inline bool			elimChecks();

	//inline bool			varElim();
	//inline bool			clsCheckBWSubsumed(const ClsPointer& clp,const uint32_t max_subsum_chk);
	//inline Literal		elimVarIf(const Variable& var,list<TempCls>& new_clses, list<TempCls>& add_anyway,vector<Message>& elims);
	//inline Literal		multiply(const Variable& var, const SharedClause& cls1,const SharedClause& cls2);
	//inline Literal 		replLit(const Variable& rem, const SharedClause& cls,Literal add);
	//inline bool 		multiply(const Variable& var, const SharedClause& cls1,const SharedClause& cls2, TempCls& out);
	//inline bool 		append(const Variable& rem, const SharedClause& cls,Literal last, uint32_t cls_strt,TempCls& out);
	//bool 				replLit(const Variable& rem, const SharedClause& cls,Literal add, TempCls& out);

	//inline Literal		tmpClsShrinkAndAdd(TempCls& cls,list<TempCls>& new_clses, list<TempCls>& add_anyway, vector<SharedClause*>& subsumed,bool take_anyway);
	//inline bool			minimizeAndCreate(Literal lits[],uint32_t l_sz, SharedClause* new_cls[])
	//inline uint32_t 	getImplsOf(const Literal& lit, Literal out[]);
	//inline void			addTempClses(list<TempCls>& clses,bool learnt);

	bool				import(){return  super::import();}
	//bool				import(){return  super::import((super::sstate.s_info->max_lbd_gc*3)/2, 90);}

};






LEADERTEMP
LEADERCLASS::SolverLeader(const GlobalSettings* sets,Satistics* sta,typename super::decision_t* dec,LeaderSiblingInfo* sinf,
		const Restarts& restart_init,const ConflictsAndGC& confl, const bool glue,
		const BeyondUP beyup,const HyperBinRes hbres,const OnTheFlySubsum otfsub):
	super						(sta,dec,sinf,glue,beyup,hbres,otfsub),
	print_cons					(sets->print_console),
	conflicts					(confl),
	restart						(restart_init),
	timeout						(sets->timeout),
	assym						(super::sstate, var_flgs,unpropd_units),
	simp_last_units				(0),
	simp_scc_on_bins			(0),
	//simp_last_max_lbd			(0),
	//simp_next_earliest			(0),
	simp_next_latest			(0),
	occ_list_units_cleared		(1)
	//analysis					(super::sstate.vinf,super::watches)
{
	asrt(super::sstate.vinf.litTrue(TRUE_LIT) && super::sstate.vinf.litFalse(FALSE_LIT));
}

LEADERTEMP
LEADERCLASS* LEADERCLASS::create(
	const GlobalSettings* sets,const uint32_t grp_sz,const uint32_t t_id,ElimClses& eli,
	typename BCPT::decision_t::act_t max,typename BCPT::decision_t::act_t decay,ImportCriteria imprt,
	uint32_t rand_freq_permil,PolarityMode m,
	const Restarts& restart_init,const ConflictsAndGC& confl,const bool glue,
	const BeyondUP beyup,const HyperBinRes hbres,const OnTheFlySubsum otfsub){

	Satistics* stats 				= new Satistics();
	LeaderSiblingInfo* sinf			= new LeaderSiblingInfo(stats,grp_sz,t_id,imprt,eli);
	typename super::decision_t* dec	= new typename super::decision_t(stats,max,decay,sets->init_rand_seed,rand_freq_permil,m);

	SolverLeader* ret				=
		new SolverLeader(sets,stats,dec,sinf,restart_init,confl,glue,beyup,hbres,otfsub);
	sinf->initVarInfo(&ret->sstate.vinf);
	return ret;
}




LEADERTEMP
void	LEADERCLASS::addValidSharedClause(SharedClause* cls,const uint32_t sz){
	super::sstate.linkCls(cls,sz,cls->getAt(0),cls->getAt(1),UserMask(0));
}


LEADERTEMP
bool	LEADERCLASS::addValidUnsharedClause(const Literal& l0, const Literal& l1){
	asrt(super::decisionLevel() == 0); asrt(!super::sstate.s_info->isActive());

	const uint32_t lsz = 1 + (l1 != INVALID_LIT);

	if(lsz == 1){return super::enqueue(l0,0) && propAt0(false);}


	asrt(lsz == 2);
	//if(super::binDag.containsBin(l0,l1)) return true;

	super::sstate.cls_reg.registerBin(l0,l1);
	super::sstate.wtch.addBinCls(l0,l1);
	super::sstate.stats->nbr_bin_clses++;
	return true;
}

LEADERTEMP
bool LEADERCLASS::enqueue(Literal lit){
	asrt(super::sstate.vinf.litUkwn(lit));
	while(!super::enqueue(lit,super::decisionLevel())){
		LOG3("Could not enqueue lit " << lit.toDIMACS());
		if(super::decisionLevel()==0) return super::setInvalid("Bad enqueue at level 0");
		super::backjumpKeepLevel(0);
		lit = lit.flipped();
	}
	return true;
}

LEADERTEMP
bool 	LEADERCLASS::sccReplace(){
	asrt(super::decisionLevel() == 0); sanity(super::asrtNoClsContainsReplOrElim());

	sendUnitsAt0();

	if(!import() || !propAt0(true)) return super::setInvalid("Import before scc replace");
	asrt(!super::openBCPImpls());

	super::sstate.stats->nbr_scc_exec++;
	vector<Message> rems; vector<Message> repls; list<list<Literal> > sccs;

	SCC scc(super::sstate);
	super::sstate.cls_reg.binsUpdated();//reset internal trigger to watch for new clauses
	scc.init(); uint32_t cnt = 0;

	do{	asrt(sccs.empty()); asrt(unpropd_units.empty());
		++cnt;

		scc.computeSCCs(sccs,unpropd_units,super::sstate.s_info->single() ? NULL : &rems);

		if(!propAt0(unpropd_units)){return super::setInvalid("Bad units after scc computation");}
		asrt(super::asrtAllUnitsSent());

		//do this before scc replacement is sent!!! example: if there are impls: a->b->c->a and x->a,x->b
		//if the scc is replaced by b then the edge x->b is not redundant anymore!!!
		if(!rems.empty()){
			super::sstate.s_info->sendMessages(rems);
			rems.clear();
		}

		for(list<list<Literal> >::iterator scc=sccs.begin();scc!=sccs.end();++scc){ asrt(scc->size()>1);
			list<Literal>::iterator li = scc->begin();
			Literal repl = *li;
			while(++li !=scc->end()){
				LOG1("Replace lit " << li->toDIMACS() << " by " << repl.toDIMACS());
				repls.push_back(Message::newLitRepl(*li,repl));
			}
			scc->clear();
		}
		sccs.clear();

		if(!repls.empty()){
			super::sstate.s_info->sendMessages(repls);
			if(!super::applyLitReplsAndRemoves(repls)){
				return super::setInvalid("Bad lit replace scc ");}
			repls.clear();
		}
		sanity(super::asrtNoClsContainsReplOrElim());



		if(!import() || !propAt0(true)){
			return super::setInvalid("Bad units after scc replace");}

	}while(cnt <3 && super::sstate.cls_reg.binsUpdated());

	return import() && propAt0(true);
}


LEADERTEMP
bool 	LEADERCLASS::removeLit(SharedClause* cls, uint32_t siz, const Literal& rem){
	Literal lits[siz]; uint32_t sz = 0;

	FOR_LITS_STRT(cls,lit);
		if(lit == rem || super::sstate.varReplFixOrElim(lit.toVar())){ continue;}
		asrt(sz<siz);
		lits[sz++]=lit;
	FOR_LITS_END;

	if(sz < 2) return sz != 0 && super::enqueue(lits[0],0) && propAt0(true);

	if(sz==2){
		super::sstate.addBinAndSend(lits[0],lits[1]);
	}else{
		SharedClause* ncls = SharedClause::construct(super::sstate.s_info->all_users_bits,lits,lits+sz,cls->getOrigLBD(),cls->isLearnt());
		super::sstate.linkCls(ncls,sz, ncls->getAt(0), ncls->getAt(1),cls->getUsers());
	}

	cls->setWaste();
	return true;
}



/*
 * assumes that for binary clauses (sub_bin,x) all x are marked
 * if super_cand is subsumed TRUE_LIT is returned, if nothing is subsumed then FALSE_LIT is returned
 * if one lit of super_cand can be removed by backward subsumption this lit is returned
 */

LEADERTEMP
Literal LEADERCLASS::chkSuperToBins(const Literal& sub_bin, const SharedClause* super,const bool allow_flpd){
	asrt(!super::lit_flags.isSet(sub_bin)); asrt(!super::lit_flags.isSet(sub_bin.flipped()));
	LOG2("Check bins of " << sub_bin.toDIMACS() << " allow flip " << allow_flpd);

	FOR_LITS_STRT(super,lit);
		if(super::lit_flags.isSet(lit)){
			LOG2(lit.toDIMACS() << " is set return sub");
			return TRUE_LIT;}

		if(allow_flpd && super::lit_flags.isSet(lit.flipped())){
			LOG2("flipped of " << lit.toDIMACS() << " is set");
			return lit;}

	FOR_LITS_END;

	return FALSE_LIT;
}

LEADERTEMP
void LEADERCLASS::clearFlgOf(const SharedClause* cls){
	FOR_LITS_STRT(cls,lit);
		super::lit_flags.carelessClear(lit);
	FOR_LITS_END;
}



LEADERTEMP
bool LEADERCLASS::collectSubsumed(const ClsPointer& subp){
	sanity(super::lit_flags.allCleared()); asrt(super::asrtAllUnitsSent());


	SharedClause* sub		= super::sstate.cinf.getCls(subp);
	if(sub->isWaste())		return true;

	ClsExtr& sube			= super::sstate.cls_reg.cls_extr.get(subp);
	const uint32_t sub_sz	= sube.sz; asrt(sub_sz>2);
	const uint32_t sub_sig	= sube.sig;

	LOG2("Collect subsumed of cls " << subp.idx() << " with size " << sub_sz << " and sig " << sub_sig);

	Literal least	= TRUE_LIT;
	uint32_t min 	= max32;

	//find variable with least occurrences
	FOR_LITS_STRT(sub,lit); asrt(!super::lit_flags.isSet(lit));
		if(super::sstate.vinf.litKnwn(lit)){
			clearFlgOf(sub);

			if(super::sstate.vinf.litTrue(lit)){
				sub->setWaste(); return true;
			}else{
				return removeLit(sub,sub_sz,lit);
			}
		}

		super::lit_flags	.setFlag(lit);
		uint32_t occ_sz = super::sstate.cls_reg.occsCount(lit.toVar());
		if(occ_sz<min)	{min=occ_sz;least=lit;}
	FOR_LITS_END;

	vector<OccCls> occs;
	super::sstate.cls_reg.copyOccsTo(least.toVar(),occs);
	bool res = true;

	for(uint32_t oi=0;res && oi != occs.size();++oi){
		ClsPointer supp 	= occs[oi].getClsP();
		if(supp==subp) 		continue;

		SharedClause* sup	= super::sstate.cinf.getCls(supp);
		if(sup->isWaste())  continue;

		ClsExtr& supe		= super::sstate.cls_reg.cls_extr.get(supp);
		if(supe.cannotBeSubsumedBy(sub_sz,sub_sig)) continue; //also considers bw subsumption

		Literal rem			= chkSuperOrEqToMarked(sup,sub_sz);

		if(rem==FALSE_LIT)	continue; // no subsumption

		if(rem==TRUE_LIT){
			super::sstate.stats->nbr_cls_subsume++;
			if(sup->isFix()){
				sub 		= super::sstate.ensureClsFix(subp,sub_sz);}
			sup				->setWaste();
		}else{
			super::sstate.stats->nbr_lits_subsume++;
			res = removeLit(sup,supe.sz,rem);
		}
	}

	clearFlgOf(sub);
	return res;
}



LEADERTEMP
Literal LEADERCLASS::chkSuperOrEqToMarked(SharedClause* sup,uint32_t sub_mrkd){

	Literal flp 	= TRUE_LIT;

	FOR_LITS_STRT(sup,lit);
		if(super::lit_flags.isSet(lit)){
			if(--sub_mrkd==0){ 		return flp;}

		}else if(super::lit_flags.isSet(lit.flipped())){
			if(flp != TRUE_LIT){	return FALSE_LIT;}
			if(--sub_mrkd==0){ 		return lit;}
			flp = lit;
		}
	FOR_LITS_END;

	asrt(sub_mrkd);
	return FALSE_LIT;
}



LEADERTEMP
bool 	LEADERCLASS::collectBinSubsumed(const Literal& lit, vector<OccCls>& occs){
	asrt(super::sstate.vals.litUkwn(lit)); sanity(super::lit_flags.allCleared()); asrt(super::some_lits.empty());
	LOG2("Collect bin subsumed start with " << lit.toDIMACS());

	if(occs.empty()) return true;

	uint32_t bin_msk = 0; Literal next = lit; uint32_t q = 0; bool res = true;
	//mark all binary partners of lit
	while(true){
		Iter<Literal> it = super::sstate.wtch.getBinsOf(next);
		while(it.getNext(next)){
			if(super::lit_flags.setFlagNew(next)){
				super::some_lits.push_back(next);
				bin_msk |= SimplyFire::getHash(next);}
		}
		if(q>=super::some_lits.size()) break;
		next = super::some_lits[q++].flipped();
	}

	if(super::lit_flags.isSet(lit)){LOG1("Unit found while collectBinSubsumed: " << lit.toDIMACS());
		unpropd_units.push_back(lit);
		goto CLEAN_BINS;}

	super::lit_flags.clear(lit.flipped()); //just delete tautological clauses

	{const bool pol = lit.isPos();
	for(uint32_t i=0;res && i!=occs.size();++i){ //copy of occs to avoid concurrent modification probs
		ClsPointer clp 		= occs[i].getClsP();
		SharedClause* cls	= super::sstate.cinf.getCls(clp);
		if(cls->isWaste())	continue;

		ClsExtr& ce			= super::sstate.cls_reg.cls_extr.get(clp);
		if(!ce.mayContainOneOf(bin_msk)){	continue;}

		const bool right_pol= (occs[i].isPosOcc() == pol);

		Literal rem			= chkSuperToBins(lit,cls,right_pol);
		if(rem==FALSE_LIT)	continue; // no subsumption

		if(rem==TRUE_LIT){
			if(right_pol){
				super::sstate.stats->nbr_cls_bin_subsume++;
				cls->setWaste();continue;}
			rem = lit.flipped(); LOG2("Flip lit to remove in Bin subsum check");
		}

		super::sstate.stats->nbr_lits_subsume++;
		res = removeLit(cls,ce.sz,rem);
	}}


	CLEAN_BINS:;
	for(uint32_t q=super::some_lits.size();q--;){
		super::lit_flags.carelessClear(super::some_lits[q]);} asrt(super::lit_flags.allCleared());
	super::some_lits.clear();
	return res;
}



LEADERTEMP
bool 	LEADERCLASS::subsumChkBins(uint32_t& cnt){
	Literal lit; cnt = 0; vector<OccCls> occs;
	while(super::sstate.cls_reg.dequeueSubLit(lit)){
		if(super::sstate.varReplFixOrElim(lit.toVar())) continue;

		++cnt;
		occs.clear();
		super::sstate.cls_reg.copyOccsTo(lit.toVar(),occs);

		if(!collectBinSubsumed(lit,occs)){
			return super::setInvalid("After Bin subsume check");}

		if(!propAt0(unpropd_units) || !import() || !propAt0(true)){
			return super::setInvalid("After Prop of Bin subsume check");}
	}

	return true;
}

LEADERTEMP
bool 	LEADERCLASS::subsumChks(){
	sendUnitsAt0();

	if(!import() || !propAt0(true)){ return super::setInvalid("Import before subsum checks");}
	asrt(!super::openBCPImpls());
	uint32_t fix = 0; uint32_t stop = super::sstate.cls_reg.clsQueueSize();
	uint32_t scount = 0;

	ClsPointer clp; uint32_t did_smthng;
	do{
		if(!subsumChkBins(did_smthng)){
			return super::setInvalid("After Bin subsume check");}


		if(super::sstate.cls_reg.dequeueSubCls(clp)){
			did_smthng = 1;

			if(!super::sstate.cinf.validNoWaste(clp)) continue;

			if(!collectSubsumed(clp)){
				return super::setInvalid("After subsumption check");}

			while(fix<super::assign_stack.size()){
				assym.setReasonForFix(super::assign_stack[fix++]);
			}

			if(!super::sstate.cinf.validNoWaste(clp)) continue;
			assym.shrinkyWinky(clp);

			if(!propAt0(unpropd_units) || !import() || !propAt0(true)){
				return super::setInvalid("After assym check");}
		}

	}while(did_smthng && restart.shallRun() && ++scount < stop);

	return import() && propAt0(true);
}


LEADERTEMP
bool	LEADERCLASS::mrkFirstFactor(const ClsBin& cb,const Literal& skip, vector<Literal>& out){
	asrt(super::lit_flags.allCleared());
	if(cb.isBin()){
		Literal lit = cb.getLit(); asrt(lit != skip); asrt(!super::sstate.vinf.litFalse(lit));
		if(super::sstate.vinf.litTrue(lit)) return true;
		out.push_back(lit);
		super::lit_flags.setFlag(lit);
		return false;
	}

	SharedClause* cls = super::sstate.cinf.getCls(cb.getClp());
	FOR_LITS_STRT(cls,lit);
		if(lit==skip || super::sstate.vinf.litFalse(lit)) continue;
		if(super::sstate.vinf.litTrue(lit)) return true;
		out.push_back(lit);
		super::lit_flags.setFlag(lit);
	FOR_LITS_END;
	return false;
}


LEADERTEMP
bool	LEADERCLASS::multiplyWithMrkd(const ClsBin& cb,vector<Literal>& out){

	if(cb.isBin()){
		Literal lit = cb.getLit();
		if(super::sstate.vinf.litTrue(lit) || super::lit_flags.isSet(lit.flipped())) return true;
		if(!super::lit_flags.isSet(lit) && !super::sstate.vinf.litFalse(lit)){out.push_back(lit);}
		return false;
	}

	SharedClause* cls = super::sstate.cinf.getCls(cb.getClp());
	FOR_LITS_STRT(cls,lit);
		if(	super::sstate.vinf.litTrue(lit) || 	super::lit_flags.isSet(lit.flipped())){	return true;}
		if(!super::lit_flags.isSet(lit) && !super::sstate.vinf.litFalse(lit)){out.push_back(lit);}
	FOR_LITS_END;
	return false;
}


LEADERTEMP
void LEADERCLASS::elimClses(ElimClses& eli, const Literal& lit, vector<ClsBin>& fakt){
	for(uint32_t i=fakt.size();i--;){
		if(fakt[i].isBin()) {
			eli.push_back(lit,fakt[i].getLit());
		}else{
			ClsPointer clp 		= fakt[i].getClp();
			ClsExtr& ce			= super::sstate.cls_reg.cls_extr.get(clp);
			SharedClause* cls 	= super::sstate.cinf.getCls(clp);
			eli.push_back(lit,cls,ce.sz);
			super::sstate.cls_reg.cls_extr.get(clp).eli_soon = 1;
		}
	}
}


//FALSE_TrBool means unsat, TRUE_TrBool means handled successfully (elim or not)
//UNKNOWN_TrBool means do again
LEADERTEMP
TrBool LEADERCLASS::elimVarIf(const Variable& var, vector<Message>& eli_mess){ asrt(super::lit_flags.allCleared());

	if(super::sstate.varReplFixOrElim(var)) return TRUE_TrBool;

	uint32_t fix_cls = super::sstate.cls_reg.cleanWasteOfVarCountFix(var);
	if(fix_cls > MAX_PRODUCT) return TRUE_TrBool;

	//import clauses since wasted ones were just deleted (may be replaced by other thread)
	if(!import() || !propAt0(true)){
		return TrBool(super::setInvalid("import before elim"));}

	if(super::sstate.varReplFixOrElim(var)) return TRUE_TrBool;

	vector<ClsBin> split[2];
	uint32_t lits	= super::sstate.cls_reg.splitFixClses(var,split,MAX_LITS,MAX_FACT);

	if(lits > MAX_LITS) return TRUE_TrBool;

	const bool pos_sec		= split[0].size() < split[1].size();
	vector<ClsBin>& fakt1	= split[!pos_sec];
	vector<ClsBin>& fakt2	= split[pos_sec]; asrt(fakt1.size() <= fakt2.size());
	const Literal sec_lit	= var.toLit(pos_sec);
	const Literal frs_lit	= sec_lit.flipped();

	vector<Literal> f1_lits; vector<std::pair<SharedClause*,uint32_t> > prods;
	TrBool ret = TRUE_TrBool;

	for(uint32_t f1 = 0;f1 != fakt1.size();++f1){

		//mark lits of first factor
		asrt(f1_lits.empty());
		bool taut1	= mrkFirstFactor(fakt1[f1],frs_lit, f1_lits);
		asrt(!super::lit_flags.isSet(var.toNegLit())); asrt(!super::lit_flags.isSet(var.toPosLit()));

		const uint32_t f1_sz = 	f1_lits.size();
		if(taut1){				goto NEXT_FAKT;}

		super::lit_flags.setFlag(sec_lit); asrt(f1_sz > 0);
		for(uint32_t f2=0;f2 != fakt2.size();++f2){
			f1_lits		.resize(f1_sz);
			bool taut2 = multiplyWithMrkd(fakt2[f2],f1_lits);

			if(taut2)	continue;

			const uint32_t psz = f1_lits.size();

			if(psz == 1){
				super::lit_flags.carelessClear(f1_lits[0]);
				ret = propAt0(f1_lits) ? UNKNOWN_TrBool
						: TrBool(super::setInvalid("Var elimination caused unit"));
				goto DONT_ELIM;}


			if(psz == 2){
				super::sstate.addBinAndSend(f1_lits[0],f1_lits[1]);
			}else if(prods.size() >= MAX_PRODUCT){
				goto DONT_ELIM;
			}else{
				SharedClause* ncls = SharedClause::construct(super::sstate.s_info->all_users_bits,f1_lits.begin(),f1_lits.end(),0,false);
				prods.push_back( std::make_pair(ncls,f1_lits.size()));
			}
		}

		NEXT_FAKT:;
		for(uint32_t i=f1_sz;i--;){super::lit_flags.carelessClear(f1_lits[i]);}
		super::lit_flags.carelessClear(sec_lit); asrt(super::lit_flags.allCleared());
		f1_lits.clear();
	}
	asrt(super::lit_flags.allCleared());

	LOG2("Eliminate var " << var.toDIMACS() << " now. Create "<<prods.size());


	//keep old clauses in elim stack
	{ElimClses& eli = super::sstate.s_info->elim;
	elimClses(eli,frs_lit,fakt1);
	elimClses(eli,sec_lit,fakt2);
	eli.push_back(var); 	LOG2("Keep Eliminated clauses " << (fakt1.size() + fakt2.size()));

	//send new clauses
	for(uint32_t i=prods.size();i--;){
		SharedClause* cls 	= prods[i].first;
		super::sstate.linkCls(cls,prods[i].second,cls->getAt(0),cls->getAt(1),cls->getUsers());
	}

	//send good news about var elimination
	eli_mess.push_back(Message::newVarElim(var));
	super::sstate.vinf	.setElimSoon(var);

	return TRUE_TrBool;



	}DONT_ELIM:;

	for(uint32_t i=f1_lits.size();i--;){super::lit_flags.carelessClear(f1_lits[i]);}
	super::lit_flags.carelessClear(sec_lit); asrt(super::lit_flags.allCleared());

	for(uint32_t i=prods.size();i--;){
		SharedClause::evilDestruct(prods[i].first);}

	return ret;
}


LEADERTEMP
bool LEADERCLASS::elimChecks(){
	asrt(super::asrtAllUnitsSent());

	vector<Message> messages;
	for(uint32_t i=0;i!=var_flgs.size();){
		Variable var(i); 	asrt(super::asrtAllUnitsSent());

		if(super::sstate.varReplFixOrElim(var)) {++i; continue;}

		TrBool el = elimVarIf(var,messages);
		if(el.isFalse()){
			return super::setInvalid("Bad var elim");}

		i += el.isTrue();
	}

	if(messages.empty()){ return true;}

	super::sstate.s_info->sendMessages(messages);
	if(!super::applyLitReplsAndRemoves(messages)){
		return super::setInvalid("Bad lit replace scc ");}

	super::clearWatches(max32);

	return import() && propAt0(true);
}










LEADERTEMP
void	LEADERCLASS::stopASAP(){
	super::sstate.stats->nbr_confl = conflicts.count();
	timeout = 0;
	restart	.stopASAP();
}


LEADERTEMP
void LEADERCLASS::paraInit(){
	uint32_t clses 		= super::sstate.stats->nAllClauses();
	conflicts			.setNumberClauses(clses);
	super::units_sent 	= super::assign_stack.size();
	super::decision		->validateNbrVariables(super::sstate.stats->nVariables());
	LOG1("Para Init finished");
	//super::binDag	.initComps();
}

LEADERTEMP
TrBool 	LEADERCLASS::isSat(){
	LOG1("random freq " << super::decision->getRandFreq());
	//LOG_BASIC("polarity mode " << m);
	LOG1("Timeout " << timeout);
	LOG1("restart init " << restart.getInitialNumber());
	LOG1("beyond Unit Prop" << super::bey_up);
	LOG1("bump_glue" << super::bump_glue);

	if(!super::ok())return FALSE_TrBool;
	paraInit(); asrt(!super::openBCPImpls());
	super::sstate.stats->init_time = Satistics::cpuWallTime();

	TrBool  status	= UNKNOWN_TrBool;

	while(Satistics::cpuWallTime() <= timeout){

		LOG_COUNT;
		LOG2("CPU User Time: "<< cpuUsrTime() << " [sec]");

		++super::sstate.stats->nbr_starts;
		asrt(!super::openBCPImpls());
		status		= search();

		if(print_cons){super::sstate.s_info->printParaProgress();}

		if(status.isKnown()) break;

		super::backjumpKeepLevel(0);
		if(!import() || !propAt0(true)){
			return TrBool(super::setInvalid("import in main loop"));}
	}

	return status;
}




LEADERTEMP
TrBool LEADERCLASS::search(){
	asrt(super::decisionLevel()==0);	asrt(super::ok()); //asrt(!super::openBCPImpls());

	restart.nextStart();

	//CONSOLE("c start search" << endl);
	TrBool		result		= UNKNOWN_TrBool;
	uint32_t	lrnts_alw;

	while(restart.shallRun()){ //|| super::decisionLevel() == 0 ){ -> can cause endless loops if conflicts set to max by other solving thread

		LOG3("New Round start bcp with conflicts " << conflicts.count() );
		int32_t bcp_res	= super::bcpLong();
		if(super::bey_up>NoBUP && bcp_res == super::BCP_SUCCESS){
			bcp_res 	= super::bcpBeyondUP();
			if(bcp_res == super::BCP_SUCCESS && super::openBCPImpls()) continue;
		}

		if(bcp_res == super::BCP_SUCCESS){		//there is no conflict
			LOG3("Prop no conflict at level " << super::decisionLevel());

			lrnts_alw = conflicts.shallDoGC(super::sstate.stats->nLearntLongClses(),super::sstate.stats->nLits());
			if(lrnts_alw != max32){
				if(!import()) 				return TrBool(super::setInvalid("bad longs import"));
				if(super::openBCPImpls())	continue;
				super::clsGC(lrnts_alw);
				if(!import()) 				return TrBool(super::setInvalid("bad longs import"));
				conflicts.didGC(			super::sstate.stats->nLearntLongClses());
			}

			//if(!importShorts()) 								return FALSE_TrBool;
			if(super::openBCPImpls()||super::openCompImpls())	continue;

			if (super::decisionLevel() == 0){
				result = simplify();
				if(result.isKnown()) return result;
				asrt(!super::openBCPImpls());
				if(super::openCompImpls())	continue;
			}


			Variable next_v	= super::decision->decideVar();

			//Literal next		= super::decision->decide();
			//if(next == INVALID_LIT){

			if (next_v == INVALID_VAR)	{
				LOG1("All variables assigned");	return TRUE_TrBool;}

			Literal next	= super::sstate.s_info->getMayorValue(next_v);
			asrt(!super::sstate.vinf.litRepl(next)); asrt(!super::sstate.vinf.varElim(next_v));

			// Increase decision level and enqueue 'next'
			asrt(super::sstate.vinf.litUkwn(next));
			super::newDecisionLevel();
			if(!enqueue(next)) return TrBool(super::setInvalid("Contradicting decision"));

		}else{						//there is a conflict
			conflicts.incConflict();

			LOG3("Prop caused "<<conflicts.count() <<". conflict at level " << super::decisionLevel());
			if(!super::handleConflict(bcp_res)) {
				return TrBool(super::setInvalid("Conflict at level 0")); }

			if(restart.incConflictBreak(super::confl_levs)) break;
		}
	}

	//maximal number of conflicts reached
	//CONSOLE("c end search" << endl);
	return UNKNOWN_TrBool;
}




LEADERTEMP
TrBool LEADERCLASS::simplify( ){

	//CONSOLE("c start simplify" << endl);
	LOG3("May be start simplify");
	asrt(super::decisionLevel() == 0);
	if (!super::ok() || !import() || !propAt0(true)){
		return TrBool(super::setInvalid("Bad Simplification"));}


	if(!sccReplace()){
		return TrBool(super::setInvalid("Bad SCC Replace"));}


	if(!subsumChks()){
		return TrBool(super::setInvalid("Bad Subsum checks"));}

	if(!sccReplace()){
		return TrBool(super::setInvalid("Bad SCC Replace after subsum check"));}


	super::clearWatches(max32);
	if(!import() || !propAt0(true)) {
			return TrBool(super::setInvalid("Bad import at simplify"));}

	asrt(!super::openBCPImpls());

	if(!elimChecks()){
		return TrBool(super::setInvalid("Bad var elimination"));}


	if(!sccReplace()){
		return TrBool(super::setInvalid("Bad SCC Replace after var elim"));}

	TIME_OFF(super::sstate.stats->simp_time);
	return UNKNOWN_TrBool;
}



LEADERTEMP
Literal 	LEADERCLASS::validateCls(vector<Literal>& lits){

	asrt(super::decisionLevel() == 0); LOG3("Validate cls with lits " << lits.size());

	if (!super::ok() || lits.empty()){ return FALSE_LIT;}

	// Check if clause is satisfied and remove false/duplicate literals:
	std::sort(lits.begin(),lits.end());
	Literal max = lits.back(); asrt(max.idx()>= lits.front().idx());
	reserveForVariables(lits.back().toVar().idx()+1);

	Literal	last(INVALID_LIT);
	vector<Literal>::iterator   to	(lits.begin());
	for(uint32_t i = 0; i < lits.size();++i){
		LOG3("Check lit" << lits[i].toDIMACS());
		if(super::sstate.vinf.litTrue(lits[i]) || lits[i] == last.flipped()){ LOG3("Cls is taut");
			lits.clear(); return TRUE_LIT;}

		if(super::sstate.vinf.litUkwn(lits[i]) && lits[i] != last){
			*to++ = last = lits[i];
			//super::decision->registerVar(last.toVar());
		}
	}
	lits.erase(to,lits.end());LOG3("Final clause size " << lits.size());
	asrt(lits.size() < 2 || (lits[0] != lits[1]));
	return max;
}


LEADERTEMP
void LEADERCLASS::cleanClause(SharedClause* cls){
	if(cls->isWaste()||cls->isSoonWaste())	return;
	vector<Literal> lits;

	FOR_LITS_STRT(cls,lit);
		if(super::sstate.vinf.litFix(lit)){
			if(super::sstate.vinf.litFalse(lit)) continue; //next Literal
			//cls->setWaste("Contains true lit ", lit.toDIMACS());return;
			cls->setWaste();return;
		}
		lits.push_back(lit);
	FOR_LITS_END;
	asrt(lits.size()>1);

	if(lits.size()==2){
		super::sstate.addBinAndSend(lits[0],lits[1]);
	}else{
		SharedClause* new_cls 	= SharedClause::construct(super::sstate.s_info->all_users_bits,
									lits.begin(),lits.end(),cls->getOrigLBD(),cls->isLearnt()); asrt(cls->getUsers().hasAnyUser());
		super::sstate.linkCls(new_cls,lits.size(),lits[0],lits[1],cls->getUsers());
	}
	//cls->setWaste("Constructed new clause for repl size ", lits.size() );
	cls->setWaste();
}

LEADERTEMP
bool LEADERCLASS::propAt0(const bool send){
	asrt(super::decisionLevel()==0);
	while(super::openBCPImpls() || super::openCompImpls()){
		if((super::bcpLong() != super::BCP_SUCCESS) || (super::bcpBeyondUP() != super::BCP_SUCCESS)) return false;
	}
	if(send){sendUnitsAt0();}
	return true;
}


LEADERTEMP
bool LEADERCLASS::propAt0(vector<Literal>& lits){ asrt(super::decisionLevel()==0);
	if(lits.empty()) return true;

	for(uint32_t i=0;i!=lits.size();++i){
		Literal lit = lits[i]; LOG1("Prop unit after simpl " << lit.toDIMACS());
		if(super::sstate.vinf.litTrue(lit)) continue;
		if(super::sstate.vinf.litFalse(lit) || !super::enqueue(lit,0)){
			return super::setInvalid("Bad unit after simpl");}
	}
	lits.clear();
	return propAt0(true);
}



LEADERTEMP
void LEADERCLASS::sendUnitsAt0(){
	asrt(super::decisionLevel()==0); //asrt(!super::openBCPImpls());
	super::sendUnits(); OccCls oc(ClsPointer(0),false);
	uint32_t to = super::assign_stack.size();
	LOG2("clear occs after units sent");
	while(occ_list_units_cleared < to){
		Variable var = super::assign_stack[occ_list_units_cleared++].toVar();
		Iter<OccCls> iter = super::sstate.cls_reg.getOccs(var);

		while(iter.getNext(oc)){
			cleanClause(super::sstate.cinf.get(oc.getClsP()).getCls());
		}
		super::sstate.cls_reg.delOccs(var);
	}
}

/*
LEADERTEMP
bool LEADERCLASS::checkBinBWSubsumed(TempCls& cls){
	asrt(super::decisionLevel()==0); asrt(cls.sorted());
	Literal other; uint32_t keep = 0; uint32_t cl_sz = cls.lits.size(); uint32_t sig  = 0;
	//cout << "start checkBinBWSubsumed with " << cl_sz << " lits"<<endl;

	for(uint32_t p=0;p!= cl_sz;++p){
		const Literal& lit = cls.lits[p]; asrt(keep<=p);
		if(super::sstate.vinf.litKnwn(lit)){
			if(super::sstate.vinf.litTrue(lit)) return true;
			continue;}

		//check if cls is subsumed by any binary clause
		if(p < cl_sz){
			ImplIter iter = super::watches.getImplIter(lit.flipped()); //subsumption with lit
			while(iter.getNext(other)){
				if(other > lit && cls.containsLit(other,p,cl_sz)){ return true;} //cls is subsumed by binary (lit,other)
			}
		}


		//check if literal can be removed from clause
		ImplIter ii = super::watches.getImplIter(lit); //backward subsumption with lit
		while(ii.getNext(other)){
			if(lit < other){ if(cls.containsLit(other,p,cl_sz)){
				//cout << " skip lit forward " << lit.toDIMACS() << " due to " << other.toDIMACS() <<endl;
				goto CONTINUE_NEXT_LIT;}
			}else if(cls.containsLit(other,0,keep)){
				//cout << " skip lit back " << lit.toDIMACS() << " due to " << other.toDIMACS() <<endl;
				goto CONTINUE_NEXT_LIT;
			}
		}

		cls.lits[keep++]	= lit;
		sig					|= SharedClause::hashVar(lit);
		CONTINUE_NEXT_LIT: ;
	}

	if(keep!=cl_sz){
		cls.lits	.resize(keep);
		cls.sig		= sig;}
	asrt(cls.sorted());

	return false;
}

LEADERTEMP
uint32_t LEADERCLASS::checkBinBWSubsumed(const SharedClause& cls, Literal cls_out[]){
	asrt(super::decisionLevel()==0);
	const uint32_t csz 				= cls.size();
	//const bool	check_normal_sub	= cls.isLearnt() || !super::cls_registry.isCompleteForFix();
	Literal other; uint32_t p = 0; uint32_t o_sz = 0;

	while(p!= csz){
		const Literal& lit = cls[p++];
		if(super::sstate.vinf.litFix(lit)){
			if(super::sstate.vinf.litFalse(lit)){continue;}
			return 0;}


		//check if cls is subsumed by any binary clause
		if(p < csz){
			ImplIter iter = super::watches.getImplIter(lit.flipped()); //subsumption with lit
			while(iter.getNext(other)){
				if(other > lit && cls.containsLit(other,p)){ return 0;} //cls is subsumed by binary (lit,other)
			}
		}


		//check if literal can be removed from clause
		ImplIter ii = super::watches.getImplIter(lit); //backward subsumption with lit
		while(ii.getNext(other)){
			if(lit < other){ if(cls.containsLit(other,p)){
				goto CONTINUE_NEXT_LIT;}
			}else if(bin_search(cls_out,0,o_sz,other) != max32){
				goto CONTINUE_NEXT_LIT;
			}
		}

		cls_out[o_sz++] = lit;
		CONTINUE_NEXT_LIT: ;
	}
	asrt(o_sz!=0);
	return o_sz;
}



LEADERTEMP
bool LEADERCLASS::clsCheckBWSubsumed(const ClsPointer& clp,const uint32_t max_subsum_chk){
	asrt(super::c_info.getCls(clp) != NULL); asrt(super::decisionLevel()==0);

	SharedClause& cls 		= *super::c_info.getCls(clp);
	if(cls.isWaste() || cls.isSoonWaste() || cls.getLBD() > max_subsum_chk)		return true;

	Literal	lits[cls.size()];  vector<SharedClause*> other_subsumed;
	SharedClause* subsuming = NULL;
	uint32_t new_sz = checkBinBWSubsumed(cls,lits);
	if(new_sz==0){
		++super::sstate.stats	->nbr_cls_bin_subsume;
	}else{

		new_sz 		= super::cls_registry.removeBWSubsumed(clp,lits, new_sz, subsuming,other_subsumed);

		//no reduction made
		if(new_sz == cls.size()) {asrt(other_subsumed.empty()); return true;}

		//check if all clauses that will be deleted are learnt
		bool all_super_learnt = cls.isLearnt(); uint32_t lbd = cls.getLBD(); uint16_t super_mask(super::sstate.s_info->user_bit.bit_mask);
		for(uint32_t i=other_subsumed.size();all_super_learnt && i--;){
			SharedClause* os	= other_subsumed[i];
			set_min_int(lbd,os->getLBD());
			all_super_learnt 	&= os->isLearnt();
			super_mask			|= os->getUsers().mask;
		}

		if(new_sz == 0){ //subsumed
			if(all_super_learnt || subsuming->isFix()){
				subsuming 				= NULL; //don't mark any as waste
			}else{
				SharedClause* new_cls	= SharedClause::copyFrom(super::sstate.s_info->all_users_bits,false,subsuming);
				super::linkCls(new_cls,(*new_cls)[0],(*new_cls)[1],true,super::sstate.s_info->all_users_bits,SubsumChk());
				LOG3("Create new clause " << (uint64_t)new_cls << " due to subsumption of fix clause from learnt" );
			}
			super::sstate.stats->nbr_cls_subsume	+= (1+other_subsumed.size());
		}else{
			super::sstate.stats->nbr_lits_subsume 	+= (cls.size()-new_sz);
			subsuming 						= NULL;
			if(new_sz > 2){
				SharedClause* new_cls 	= SharedClause::construct(super::sstate.s_info->all_users_bits,
												lits,lits+new_sz,lbd,all_super_learnt); asrt(super_mask != 0);
				super::linkCls(new_cls,lits[0],lits[1],true,UserMask(super_mask),SubsumChk());
				LOG3("Create new clause " << (uint64_t)new_cls << " due to lit replace " );
			}else if(new_sz == 2){
				if(!super::watches.containsBin(lits[0],lits[1])){
					super::addBinAndSend(lits[0],lits[1]);}
				LOG3("Create new binary clause " << lits[0].toDIMACS() << "," << lits[1].toDIMACS()<< " due to lit replace " );
			}else{
				asrt(new_sz==1); LOG3("Create new unit clause " << lits[0].toDIMACS()<< " due to lit replace " );
				if(!super::enqueue(lits[0],0) || !propAt0()){
					return super::setInvalid("createReplaced produced conflict");}
				//sendUnitsAt0();
			}
		}
	}asrt(!super::openBCPImpls());


	sendUnitsAt0();
	LOG2("Set Waste clauses thanks to subsumption" );
	cls.setWaste();//cls.setWaste("is subsumed in bw sub");
	for(uint32_t i=other_subsumed.size();i--;){other_subsumed[i]->setWaste();}
	if(subsuming != NULL){subsuming->setWaste();}
	return true;
}


LEADERTEMP
uint32_t LEADERCLASS::getImplsOf(const Literal& l, Literal out[]){
	uint32_t sz = 0; Literal lit;
	ImplIter it = super::watches.getImplIter(l);
	while(it.getNext(lit)){ asrt(!super::sstate.vinf.varEliminated(lit.toVar()));
		if(!super::sstate.vinf.litFix(lit) && !super::cls_registry.varSoonElim(lit.toVar())){
			out[sz++]	= lit;}
	}
	return sz;
}


#define add_if(a) 	asrt(!super::sstate.vinf.litKnwn(a));\
					if(a==last.flipped()){ return true;} \
					if(last!=a && a.toVar() != rem){out.addLit(last=a);}


LEADERTEMP
bool LEADERCLASS::multiply(const Variable& rem, const SharedClause& cls1,const SharedClause& cls2, TempCls& out){
	uint32_t idx1 	= 0; uint32_t idx2 = 0; asrt(out.sig==0);
	Literal last 	= INVALID_LIT; //Literal a = cls1[0]; Literal b = cls2[0];

	//cout << "start multiply" << endl;
	while(idx1<cls1.size() && idx2<cls2.size()){
		Literal next = cls1[idx1]<cls2[idx2] ? cls1[idx1++]: cls2[idx2++];
		add_if(next)
		asrt(!super::cls_registry.varSoonElim(next.toVar())); asrt(!super::sstate.vinf.varEliminated(next.toVar()));
	}
	asrt(out.sorted());
	return append(rem,cls1,last,idx1,out) || append(rem,cls2,last,idx2,out);
}


 // return the size of the factor clause. if size == max32 the multiplication is tautological

LEADERTEMP
bool 	LEADERCLASS::replLit(const Variable& rem, const SharedClause& cls,Literal add, TempCls& out){
	asrt(super::decisionLevel()==0); asrt(add != INVALID_LIT);asrt(!super::cls_registry.varSoonElim(add.toVar()));
	//cout << "start repl Lit with add "<< add.toDIMACS() << endl;

	Literal last = INVALID_LIT;
	for(uint32_t idx=0;idx<cls.size();){
		Literal a = cls[idx];
		if(add<a){a=add;add=INVALID_LIT;}else{++idx;}
		add_if(a)
		asrt(!super::sstate.vinf.varEliminated(a.toVar())); asrt(!super::cls_registry.varSoonElim(a.toVar()));
	}
	if(add != INVALID_LIT){add_if(add)}
	//cout << "finish repl Lit" << endl;
	asrt(out.sorted());
	return false;
}

LEADERTEMP
bool 	LEADERCLASS::append(const Variable& rem, const SharedClause& cls,Literal last, uint32_t cls_strt,TempCls& out){
	asrt(super::decisionLevel()==0);//cout << "start append " << endl;

	while(cls_strt< cls.size()){
		Literal a = cls[cls_strt++]; asrt(!super::cls_registry.varSoonElim(a.toVar())); asrt(!super::sstate.vinf.varEliminated(a.toVar()));
		add_if(a)}

	//cout << "finish append " << endl;
	asrt(out.sorted());
	return false;
}





LEADERTEMP
Literal	LEADERCLASS::tmpClsShrinkAndAdd(TempCls& cls,list<TempCls>& new_clses, list<TempCls>& add_anyway, vector<SharedClause*>& subsumed,bool take_anyway){
	asrt(super::decisionLevel()==0);
	SharedClause* subsuming = NULL; uint32_t sss = subsumed.size();

	if(checkBinBWSubsumed(cls)) return TRUE_LIT;
	uint32_t cl_sz = cls.lits.size();
	if(cl_sz==1)				return cls.lits[0];

	cl_sz 	= super::cls_registry.removeBWSubsumed(ClsPointer(),cls.lits.begin().base(),
					cl_sz, subsuming,subsumed); asrt(cl_sz <= cls.lits.size());
	cls		.reArrange(cl_sz);	asrt(cls.sorted());

	if(cl_sz==1)				return cls.lits[0];

	take_anyway 				|= (sss != subsumed.size());
	if(cl_sz==0){//subsumed
		LOG1("Subsumed in tmpClsShrinkAndAdd");
		if(subsuming->isFix())	return TRUE_LIT;
		take_anyway	= true;
		cls = TempCls(*subsuming);
	}

	//falls nicht take_anyway nochmal gucken ob irgendeine subsumiert wird
	return Subsumption::addCheckSub((take_anyway?add_anyway:new_clses),cls);
}


LEADERTEMP
Literal	LEADERCLASS::elimVarIf(const Variable& var,list<TempCls>& new_clses, list<TempCls>& add_anyway,vector<Message>& elims){
	asrt(super::decisionLevel()==0); asrt(new_clses.empty()); asrt(add_anyway.empty()); asrt(!super::cls_registry.varSoonElim(var));

	//get binaries of negative and positive literal
	uint32_t pb_sz	= super::watches.implCount(var.toNegLit());
	Literal pos_b[pb_sz]; pb_sz = getImplsOf(var.toNegLit(),pos_b);
	uint32_t nb_sz	= super::watches.implCount(var.toPosLit());
	Literal neg_b[nb_sz]; nb_sz = getImplsOf(var.toPosLit(),neg_b);

	//early check
	if(nb_sz*pb_sz > Subsumption::max_elim_product) 	return TRUE_LIT;

	//split fix clauses into pos and neg and check if elimination is still possible
	SharedClause* pos_l[Subsumption::max_elim_product]; uint32_t pl_sz = 0;
	SharedClause* neg_l[Subsumption::max_elim_product]; uint32_t nl_sz = 0;
	if(!super::cls_registry.splitFixClses(var,pos_l,pl_sz,neg_l,nl_sz)) return TRUE_LIT;
	LOG2("For var"<< var.toDIMACS()<< ": " << pl_sz << " pos and " << nl_sz << " neg long fix clauses");

	//new limit
	const uint32_t max_new = pl_sz + pb_sz + nl_sz + nb_sz;
	Literal ret; vector<SharedClause*> subsumed;

	//cross product of binaries
	for(uint32_t i=0;i<pb_sz;++i){
		Literal a = pos_b[i]; asrt(!super::watches.isLitRepl(a)); asrt(!super::sstate.vinf.varEliminated(a.toVar()));asrt(!super::sstate.vinf.litFix(a));

		for(uint32_t j=0;j<nb_sz;++j){
			Literal b = neg_b[j]; asrt(!super::watches.isLitRepl(b)); asrt(!super::sstate.vinf.varEliminated(b.toVar()));asrt(!super::sstate.vinf.litFix(b));

			if(a==b){asrt(super::sstate.vinf.litUkwn(a));
				LOG1("Elim of " << var.toDIMACS() << " caused unit " << a.toDIMACS());
				return a;}

			if(a==b.flipped() || super::watches.containsBin(a,b)) continue;

			if(super::cls_registry.isThereAnySubsumedOfBinary(a,b) != NULL){
				LOG1("Elim of " << var.toDIMACS() << " created subsuming binary " << a.toDIMACS()<<" "<<b.toDIMACS());
				super::addBinAndSend(a,b);
			}else{
				TempCls bin(2);
				if(a<b)	{bin.addLit(a);bin.addLit(b);}else{bin.addLit(b);bin.addLit(a);}
				ret = tmpClsShrinkAndAdd(bin,new_clses,add_anyway,subsumed,false);
				if(ret != TRUE_LIT ||  new_clses.size()>max_new) return ret;
			}
		}
	}

	//multiply long clauses with positive literal of "var" with binaries
	for(uint32_t i=0;i<pb_sz;++i){
		Literal a = pos_b[i]; asrt(!super::watches.isLitRepl(a)); asrt(!super::sstate.vinf.varEliminated(a.toVar()));asrt(!super::sstate.vinf.litFix(a));
		for(uint32_t j=0;j<nl_sz;){
			SharedClause* cls = neg_l[j];	TempCls add(cls->size());
			if(replLit(var,*cls,a,add)){++j; continue;}

			bool take	= add.lits.size()<cls->size(); //resolvant subsumes cls
			asrt(!take || add.lits.size()+1==cls->size());


			ret 		= tmpClsShrinkAndAdd(add,new_clses,add_anyway,subsumed,take);
			if(ret != TRUE_LIT ||  new_clses.size()>max_new) return ret;


			if(take){neg_l[j]=neg_l[--nl_sz];LOG1("Multiplication subsumes Multiplicator");}
			else{++j;}
		}
	}

	//multiply long clauses with negative literal of "var" with binaries
	for(uint32_t i=0;i<nb_sz;++i){
		Literal a = neg_b[i]; asrt(!super::watches.isLitRepl(a)); asrt(!super::sstate.vinf.varEliminated(a.toVar()));asrt(!super::sstate.vinf.litFix(a));
		for(uint32_t j=0;j<pl_sz;){
			SharedClause* cls = pos_l[j]; 	TempCls add(cls->size());
			if(replLit(var,*cls,a,add)){++j; continue;}

			bool take	= add.lits.size()<cls->size(); //resolvant subsumes cls
			asrt(!take || add.lits.size()+1==cls->size());

			ret 		= tmpClsShrinkAndAdd(add,new_clses,add_anyway,subsumed,take);
			if(ret != TRUE_LIT ||  new_clses.size()>max_new) return ret;

			if(take){pos_l[j]=pos_l[--pl_sz];LOG1("Multiplication subsumes Multiplicator");}
			else{++j;}
		}
	}

	//multiply long clauses
	for(uint32_t i=0;i<pl_sz;){
		SharedClause* cl1 = pos_l[i]; bool sub1=false;
		for(uint32_t j=0;j<nl_sz;){
			SharedClause* cl2 = neg_l[j]; TempCls add(cl1->size()+cl2->size());
			if(multiply(var,*cl1,*cl2,add)) {++j;continue;}

			sub1 		= add.lits.size()<cl1->size();
			bool sub2 	= add.lits.size()<cl2->size();
			asrt(!sub1 || add.lits.size()+1==cl1->size());	asrt(!sub2 || add.lits.size()+1==cl2->size());

			ret = tmpClsShrinkAndAdd(add,new_clses,add_anyway,subsumed,sub1||sub2);
			if(ret != TRUE_LIT ||  new_clses.size()>max_new) return ret;

			if(sub2){neg_l[j]=neg_l[--nl_sz];LOG1("Multiplication subsumes inner Multiplicator");}
			else{++j;}
			if(sub1){ break;}
		}
		if(sub1){pos_l[i]=pos_l[--pl_sz];LOG1("Multiplication subsumes outer Multiplicator");}
		else{++i;}
	}

	LOG2("Eliminate var " << var.toDIMACS() << " now. Create "<<new_clses.size()<< " anyway: "<<add_anyway.size());

	//keep old clauses in elim stack
	ElimClses& eli = super::sstate.s_info->elim;
	Literal pos = var.toPosLit(); Literal neg = var.toNegLit();
	for(uint32_t i=pl_sz;i--;){	eli.push_back(pos,pos_l[i]);}
	for(uint32_t i=nl_sz;i--;){	eli.push_back(neg,neg_l[i]);}
	for(uint32_t i=pb_sz;i--;){	eli.push_back(pos,pos_b[i]);asrt(!super::cls_registry.varSoonElim(pos_b[i].toVar()));}
	for(uint32_t i=nb_sz;i--;){	eli.push_back(neg,neg_b[i]);asrt(!super::cls_registry.varSoonElim(neg_b[i].toVar()));}
	eli.push_back(var);
	LOG2("Keep Eliminated clauses " << (pl_sz+nl_sz+pb_sz+nb_sz));

	//send new clauses
	addTempClses(new_clses,false);
	addTempClses(add_anyway,false);

	//send good news about var elimination
	elims.push_back(Message::newVarElim(var));

	LOG1("Set soon elim var " << var.toDIMACS());
	super::cls_registry.setSoonElim(var);

	for(uint32_t i=subsumed.size();i--;){subsumed[i]->setWaste();}

	new_clses.clear();add_anyway.clear();subsumed.clear();


	return TRUE_LIT;
}

LEADERTEMP
void LEADERCLASS::addTempClses(list<TempCls>& clses,bool learnt){
	for(list<TempCls>::iterator iter = clses.begin();iter!=clses.end();++iter){
		uint32_t lsz = iter->lits.size(); asrt(lsz>1);
		Literal w0 = iter->lits[0]; Literal w1 = iter->lits[1];
		asrt(super::sstate.vinf.litUkwn(w0)); asrt(super::sstate.vinf.litUkwn(w1)); asrt(w0!=w1);
		asrt(!super::cls_registry.varSoonElim(w0.toVar())); asrt(!super::cls_registry.varSoonElim(w1.toVar()));
		if(lsz==2){
			super::addBinAndSend(w0,w1);
		}else{
			asrt(iter->sorted());
			SharedClause* cls = SharedClause::construct(super::sstate.s_info->all_users_bits,iter->lits.begin(),
							iter->lits.end(),lsz,learnt); //lbd will be set to 0 if learnt is false
			super::linkCls(cls,w0,w1,true,UserMask(super::sstate.s_info->user_bit),SubsumChk());
		}


	}
}
*/


/*
LEADERTEMP
bool LEADERCLASS::clsAnalyse(const ClsPointer& clp,const uint32_t max_subsum_chk){
	asrt(super::c_info.getCls(clp) != NULL); asrt(super::decisionLevel()==0); asrt(super::lit_flags.allCleared());

	SharedClause& cls 		= *super::c_info.getCls(clp);
	if(cls.isWaste())		return true;
	//if(cls.isRedundant()){	return importShortsAtL0() ? (SharedClause::setTrash(&cls,true),true): false;}

	//load changes of clause, apply literal replacements
	Literal	lits[cls.size()]; uint32_t l_sz = 0;
	bool change			= false;


	// mittlerweile total überflüssig!!!
	for(uint32_t i=0;i!= cls.size();++i){
		Literal l		= super::watches.findLitRepl(cls[i]);
		asrt(l != FALSE_LIT || super::sstate.vinf.litFixFalse(cls[i]));
		asrt(l != TRUE_LIT  || super::sstate.vinf.litFixTrue(cls[i]));
		if(super::sstate.vinf.litTrue(l) || super::isSet(l.flipped())){
			super::zeroFlaggs(lits,l_sz); asrt(super::lit_flags.allCleared());
			super::sendUnits(); cls.markWaste();;
			LOG_DETAIL("Clause is sat"); return true;}

		if(super::sstate.vinf.litFalse(l) || super::isSet(l)){change=true;continue;}
		super::setFlag(	lits[l_sz++]= l);
		change			|= (l!=cls[i]);
	}

	asrt(change || l_sz==cls.size());
	if(change){	std::sort(lits,lits+l_sz);}

	if(cls.getLBD() <= max_subsum_chk){
		l_sz 	= checkSubsumption(lits,l_sz,&cls,!cls.isLearnt());
	}else{LOG_MEDIUM("Omit subsum check due to high lbd value");
		super::zeroFlaggs(lits,l_sz);}
	asrt(super::lit_flags.allCleared());


	if(l_sz == max32){
		LOG_MEDIUM(" Clause subsumed " << (uint64_t)(&cls)<< " with size " <<cls.size()<<" starting with " << cls[0].toDIMACS() << " new " << lits[0].toDIMACS());
		super::sendUnits();
		cls.markWaste();
		return true;
	}

	if(l_sz == cls.size() && !change){
		LOG_MEDIUM("Clause is unchanged ");
		return true;}

	if(l_sz < 3){

		if(l_sz == 2){
			if(!super::watches.containsBin(lits[0],lits[1])){
					super::addBinAndSend(lits[0],lits[1]);}
		}else{
			asrt(l_sz==0 || super::sstate.vinf.litUkwn(lits[0]));
			if(l_sz==0 || !super::enqueue(lits[0],0) || super::bcpLong() != super::BCP_SUCCESS){
				return super::setInvalid("createReplaced produced conflict");
			}
		}

		super::sendUnits();
		LOG_MEDIUM("Set Waste clause " << ((uint64_t)&cls) << " replaced by " << l_sz);
		cls.markWaste();
		return 	true;
	}

	LOG_DETAIL("shrink clause to size " << l_sz); asrt(lits[0]!=lits[1]);
	//check new long clause
	SharedClause* new_cls 	= SharedClause::construct(super::sstate.s_info->all_users_bits,lits,lits+l_sz,
								min(cls.getLBD(),l_sz),cls.isLearnt());
	LOG_DETAIL("Create new clause " << (uint64_t)new_cls << " due to lit replace " );

	super::linkSharedCls(new_cls,lits[0],lits[1],true);

	super::sendUnits();
	LOG_DETAIL("Set Waste clause " << ((uint64_t)&cls) << " replaced by " << ((uint64_t)new_cls));
	cls.markWaste();
	return true;
}


struct SubsumeVisitor{
	BitVector& 		flags;
	bool			sub;
	SubsumeVisitor(BitVector& lit_flags):flags(lit_flags),sub(false){}
	TrBool operator()(const Literal& tgt, const Literal& src){
		if(flags.isSet(tgt.idx())){
			sub = true;	return UNKNOWN_TrBool;} //stop immediately -> clause is subsumed
		flags	.clear(tgt.flipped().idx()); //clear flag if set -> literal resolved out by bw-subsumption
		return 	TRUE_TrBool;
	}
};
*/

/**
 * Input: Sorted Array of literals who are checked to be subsumed - exactly these literals are flagged
 * Return new size of lits array (<= l_sz if lits are resolved out by bw-subsumption)
 * If return value is max32 the given set of literals is redundant
 * the clause "ignore" is ignored as subsuming clause
 * if ensure_cls is true and the subsuming cls is learnt than a copy of the subsuming cls marked as "not learnt" is constructed
 */
/*
LEADERTEMP
uint32_t 	LEADERCLASS::checkSubsumption(Literal lits[], const uint32_t& l_sz,SharedClause* ignore, const bool ensure_cls){
	uint32_t ret 			= l_sz;
	uint32_t marked_right	= 0;
	SubsumeVisitor 			visi(super::lit_flags);

	if(l_sz <= 2) {			goto CLEAR_RETURN;}
	sanity(super::asrtSortedAndFlagged(lits,l_sz));

	for(uint32_t i=l_sz;--i && !visi.sub;){ //do for all but one
		if(!super::isSet(lits[i])) 	continue;
		super::BFS(lits[i].flipped(),visi,false,false);	asrt(super::isSet(lits[i]));
	}
	super::clearBFSFlags();
	if(visi.sub){ret=max32; super::sstate.stats->nbr_cls_bin_subsume++;	goto CLEAR_RETURN;}

	for(uint32_t idx = l_sz;idx--;){
		bool this_mkd		= super::isSet(lits[idx]);
		marked_right		+= this_mkd;
		if(marked_right<3 || !this_mkd)	continue; //no clause with smallest lit lits[idx] can subsume the marked literals

		uint32_t remov		= checkSubsumed(lits[idx],marked_right,ignore); //ignore
		if(remov!=max32){	asrt(remov<marked_right);
			marked_right	-= remov; continue;}

		//ignore holds the subsuming clause
		asrt(ignore != NULL); //asrt(!ignore->isWaste());
		if(ensure_cls && ignore->isLearnt()){

			asrt(ignore->size()<=l_sz);
			//create copy of sub since other threads might just delete sub while we mark cls as waste
			SharedClause* n	= SharedClause::copyFrom(super::sstate.s_info->all_users_bits,false,ignore);
							asrt(n->size()>2); asrt(!super::sstate.vinf.litKnwn((*n)[0])); asrt(!super::sstate.vinf.litKnwn((*n)[1]));
			super::	linkSharedCls(n,(*n)[0],(*n)[1],true);
			ignore->markWaste();
			LOG_DETAIL("Create new clause " << (uint64_t)n << " as copyFrom " << (uint64_t)ignore);
		}

		ret				= max32;
		super::sstate.stats	->nbr_cls_subsume++;
		goto			CLEAR_RETURN;
	}

	asrt(marked_right != 0); asrt(ret==l_sz);
	if(marked_right == l_sz){ //clause is unchanged
		CLEAR_RETURN: ;
		for(uint32_t i=l_sz;i--;){super::zeroFlags(lits[i]);}
	}else{
		ret=0;
		for(uint32_t i=0;ret!=marked_right;++i){asrt(i<l_sz);
			if(super::isSet(lits[i])){super::clearFlag(lits[ret++]=lits[i]);}}
		asrt(ret==marked_right); asrt(ret<l_sz);
	}

	asrt(super::lit_flags.allCleared());
	super::sstate.stats->nbr_lits_subsume += (l_sz-ret);
	return ret;
}



LEADERTEMP
uint32_t LEADERCLASS::checkSubsumed(const Literal& lit, const uint32_t& lits_above, SharedClause*& dont){

	const Variable var			= lit.toVar();
	IVec<ClsPointer>& occs 		= super::cls_registry.get(var);
	WriteIter<ClsPointer> wi	= occs.getWriter();
	uint32_t removed			= 0;

	while(wi.hasNext()){
		ClsPointer& clp				= wi.getNext();
		if(!super::c_info.validPtr(clp)){
									continue;}
		SharedClause* sub			= super::c_info.getCls(clp);
									asrt(sub!= NULL); asrt(sub->size()>2); asrt(super::asrtClsSorted(*sub));
		if(sub->isWaste() || (*sub)[0].toVar()!=var){
									continue;}
		wi.keepThis(clp);
		if(sub==dont || sub->size()>lits_above) continue;

		Literal bw_subsum			= INVALID_LIT;
		uint32_t sup_cnds			= lits_above;
		for(uint32_t b=sub->size();b;--sup_cnds){
			if(b-- >sup_cnds){		goto CONTINUE_NEXT_SUB;} //too many lits in subsum candidate

			Literal s 				= (*sub)[b]; asrt(sup_cnds>0);
			if(super::isSet(s))		continue; //just decrement number of literals in super cls (sup_cnds)

			if(!super::isSet(s.flipped()) || (bw_subsum != INVALID_LIT)){
									goto CONTINUE_NEXT_SUB;}

			bw_subsum				= s.flipped();
		}

		if(bw_subsum==INVALID_LIT){ //marked lits are subsumed by clp
			wi	.keepRest();	occs	.shrinkTo(wi);
			dont= sub;			return 	max32; //indicate real subsumption
		}

		//backward subsumption
		asrt(super::isSet(bw_subsum));
		super::clearFlag(bw_subsum);
		++removed;

		if(bw_subsum==lit){ //cannot remove other lit since lit
			wi.keepRest();occs.shrinkTo(wi);
			return removed;
		}

		asrt(super::isSet(lit));
		CONTINUE_NEXT_SUB:;
	}
	occs.shrinkTo(wi);
	return removed;
}
*/


/*
LEADERTEMP
bool LEADERCLASS::checkAutarkies(){
	AutarkyGuess* guess; asrt(super::decisionLevel()==0);
	if(!propAt0()) return false;
	sendUnitsAt0();

	while((guess=super::sstate.s_info->popNextAutarkyGuess(),guess != NULL)){

		if(conflicts.isMax()) return true;

		LOG2(" next Autarky guess with size " << guess->sz);

		TrBool res = checkAutarky(guess->lits,guess->sz); asrt(super::decisionLevel()==0);


		if(res.isUnknown()) {continue;}

		if(res.isFalse()) 	return false;

		//cout << " found autarky" << endl;

		if(!propAt0() || !import() || !propAt0()) return false;
		sendUnitsAt0();
		// alten guess filtern -> vielleicht kommt nochmal was raus

	}

	return true;
}




struct MarkVisitor{
	const BinVarInfo& v_info;
	bool ok;
	MarkVisitor(const BinVarInfo& v_i):v_info(v_i),ok(true){}
	TrBool operator()(const Literal& tgt, const Literal& src){
		return v_info.isFlagged(tgt.flipped()) ? UNKNOWN_TrBool : TRUE_TrBool;}
};




LEADERTEMP
TrBool LEADERCLASS::checkAutarky(const Literal lits[], const uint32_t sz){
	asrt(super::decisionLevel()==0); sanity(super::sstate.vinf.asrtAllFlagsClear());asrt(super::tmp_literal_store.empty());
	asrt(sz>0); asrt(super::cls_registry.isCompleteForFix()); asrt(!super::openCompImpls()); asrt(!super::openBCPImpls());
	sendUnitsAt0();
	TrBool ret = UNKNOWN_TrBool;

	//mark all literals by binary implications
	MarkVisitor mark(super::sstate.vinf);
	for(uint32_t i = sz;i--;){
		Literal lit = lits[i];
		if(super::sstate.vinf.litKnwn(lit)){
			super::clearBFSFlags(); return UNKNOWN_TrBool;}

		if(	super::sstate.vinf.isFlagged(lit)
		||	super::sstate.vinf.varEliminated(lit.toVar())
		||	super::watches.isLitRepl(lit)) continue;

		super::BFS(lit,mark,true,false);
		if(!mark.ok){
			super::clearBFSFlags(); return UNKNOWN_TrBool;}
	}
	if(super::tmp_literal_store.empty()) return UNKNOWN_TrBool;

	//enqueue all clauses to be satisfied into "some_lits" and decide on which literal to assign next
	//candidates for literals are only the marked ones
	super::newDecisionLevel(); asrt(super::decisionLevel()==1);
	Literal next = super::tmp_literal_store.back(); ClsPointer clp;
	while(true){

		uint32_t chk_beg = super::queue_begin;
		//next choice
		if(!super::enqueue(next,1)){next = next.flipped(); goto FINISH_AUT_CHK;}
		next = INVALID_LIT;

		//filter clses to be satisfied
		uint32_t cp = 0; uint32_t last_strt=0; bool curr_sat = false;
		for(uint32_t i=0;i!=super::some_lits.size();++i){ asrt(last_strt<=cp);
			Literal lit = super::some_lits[i];
			if(lit == INVALID_LIT){ //end of cls
				uint32_t sz		= cp - last_strt;
				if(curr_sat){	cp = last_strt; asrt(last_strt==0|| super::some_lits[last_strt-1]==INVALID_LIT);
				}else if(sz==0)	{next = INVALID_LIT; goto FINISH_AUT_CHK;
				}else{
					if(sz==1)	{asrt(last_strt<super::some_lits.size()); next=super::some_lits[last_strt];}
								super::some_lits[cp++]=INVALID_LIT;last_strt = cp;}
				curr_sat 		= false;
			}else if(curr_sat || super::sstate.vinf.litTrue(lit)){
				curr_sat=true;
			}else if(!super::sstate.vinf.litFalse(lit)){
				super::some_lits[cp++] = lit;}
		}
		super::some_lits.resize(cp);

		//add clauses to be sat
		while(chk_beg != super::assign_stack.size()){
			Literal lit = super::assign_stack[chk_beg++];

			//check clauses with literal lit.flipped()
			OccIter cit = super::cls_registry.getOccsOf(lit.toVar());
			while(cit.getNextWithPolartiy(clp,lit.isNeg())){
				SharedClause* cls = super::c_info.getCls(clp);
				if(cls->isLearnt()) goto NEXT_CLS_CHK;

				{const uint32_t slz = super::some_lits.size();
				for(uint32_t i=cls->size();i--;){
					Literal cand = (*cls)[i];
					if(super::sstate.vinf.litTrue(cand)){ //cls is already satisfied
						super::some_lits.resize(slz); goto NEXT_CLS_CHK;}

					if(super::sstate.vinf.litUkwn(cand) && super::sstate.vinf.isFlagged(cand)){
						super::some_lits.push_back(cand); } //add candidate to sat cls
				}

				if(super::some_lits.size()==slz)	{next=INVALID_LIT; goto FINISH_AUT_CHK;} // no candidate to sat this cls was found
				if(super::some_lits.size()==slz+1)	{next=super::some_lits[slz];}

				super::some_lits.push_back(INVALID_LIT);} // end-tag for some set of literals
				NEXT_CLS_CHK:;
			}
		}

		//already know next decision (forced)
		if(next != INVALID_LIT) continue;

		//found autarky
		if(super::some_lits.empty()){
			asrt(super::decisionLevel()==1);
			uint32_t beg = super::decision_level_end[0];
			super::decision_level_end.clear();//lits already enqueued
			super::sstate.stats->nbr_autarkies += (super::assign_stack.size()-beg);

			while(beg != super::assign_stack.size()){
				super::sstate.vinf.setValueDecision(super::assign_stack[beg++],0);}

			ret = TRUE_TrBool; asrt(next==INVALID_LIT);	goto FINISH_AUT_CHK;
		}

		// besser machen
		next = super::some_lits[0];
		for(uint32_t i=1;i<super::some_lits.size();++i){
			Literal l = super::some_lits[i];
			if(l==INVALID_LIT) break;
			if(super::decision->getAct(l)<super::decision->getAct(next)){next=l;}
		}


		asrt(next != INVALID_LIT);
	}

	FINISH_AUT_CHK:;
	super::some_lits.clear();
	super::backjumpKeepLevel(0);
	super::clearBFSFlags();
	return next != INVALID_LIT ? TrBool(super::enqueue(next,0)) : ret;
}

*/


/*
LEADERTEMP
bool 	LEADERCLASS::varElim(){
	asrt(super::decisionLevel()==0); sanity(super::asrtNoClsContainsReplOrElim());
	makeOccListCompleteForFixClses();
	if(!import() || !propAt0()){
		return super::setInvalid("bad import after making occ list complete");}

	uint32_t pl_sz, nl_sz; list<TempCls> new_clses; list<TempCls> add_anyway; vector<Message> removes;
	for(uint32_t v = 1; v < super::sstate.vinf.varsCount();++v){
		Variable var(v);
		if(super::sstate.vinf.varKnwn(var) || super::watches.isVarRepl(var) || super::sstate.vinf.varEliminated(var)) continue;
		asrt(!super::cls_registry.varSoonElim(var));

		uint32_t lits	= super::cls_registry.cleanWasteOfVarGetLitsCount(var,pl_sz,nl_sz);

		//import immediately to get updates of recently wasted clauses
		if(!import() || !propAt0()){
			return super::setInvalid("Contradicting import");}

		//check if possible to eliminate
		if((lits > Subsumption::max_elim_lits)
		||(pl_sz * nl_sz > Subsumption::max_elim_product)) 	continue;

		if(occ_list_units_cleared < super::assign_stack.size()){
			//to remove all clauses with fix literals
			sendUnitsAt0();
			LOG2("Repeat vor same var in varElim");
			--v; continue; //with same literal
		}


		Literal res = elimVarIf(var,new_clses,add_anyway,removes);
		if(res != TRUE_LIT){asrt(super::sstate.vinf.litUkwn(res));
			if(!super::enqueue(res,0)||(!propAt0())){
				return super::setInvalid("Bad resolved unit in var elimination");}
			sendUnitsAt0();
			--v; asrt(!super::cls_registry.varSoonElim(var));
		}else{
			//if elimVar gave up then delete the already created clauses
			addTempClses(add_anyway,true);
		}
		new_clses.clear(); add_anyway.clear();



	}

	bool ok = true;
	if(!removes.empty()){
		LOG2("Eliminate var send messages: "<<removes.size() );
		super::sstate.s_info->sendMessages(removes);
		LOG2("Eliminate var start own removal with : "<<removes.size() );
		ok = super::applyLitReplsAndRemoves(removes);
		LOG2("Eliminate var finish with: "<<removes.size() );
	}
	sanity(super::asrtNoClsContainsReplOrElim());

	return ok && import() && propAt0();
}

*/

/*

LEADERTEMP
bool 	LEADERCLASS::failedProp(const Literal& lit){
	asrt(super::decisionLevel()==0);asrt(!super::openBCPImpls()); asrt(super::bey_up>NoBUP);
		super::newDecisionLevel(); asrt(super::decisionLevel()==1);

		//only try one polarity since binary clauses are added by hyper bin resolution within bcpLong
		bool ok = super::enqueue(lit,1);

		while(ok){
			ok 	= 	super::bcpLong() 	== super::BCP_SUCCESS &&
					super::bcpBeyondUP()== super::BCP_SUCCESS;
			if(!super::openBCPImpls()){ break;}
		}

		super::backjumpZero();

		LOG3("failed prop fini " << lit.toDIMACS() << " bin vars " << super::sstate.stats->nbr_bin_clses);
		if(ok) return true;;

		super::lemma.clear();
		super::delLevel(1);

		return super::enqueue(lit.flipped(),0) && propAt0();
}


LEADERTEMP
bool 	LEADERCLASS::failedProp(){
	asrt(super::decisionLevel() == 0); asrt(!super::openBCPImpls()); asrt(super::hb_res>NoHyperBin);
	if(	!import() || !propAt0()){return false;}

	for(uint32_t i=super::nVars();i--;){
		Variable var(i);
		if(super::sstate.vinf.varKnwn(var) || super::sstate.vinf.varEliminated(var) || super::watches.isLitRepl(var.toPosLit())) continue;

		if(conflicts.isMax()) break;

		const bool p_is_root = super::watches.isSink(var.toNegLit());
		const bool n_is_root = super::watches.isSink(var.toPosLit());
		if(p_is_root && n_is_root) continue;

		if(	(p_is_root && !failedProp(var.toPosLit()))
		||	(n_is_root && !failedProp(var.toNegLit()))) return super::setInvalid("Contradiction in FailedProp");

	}

	asrt(super::decisionLevel()==0);
	super::sendUnits();

	return propAt0();
}
*/


/*
LEADERTEMP
TrBool 	LEADERCLASS::createSharedBinCls(Literal l0, const Literal& l1,const UserID& found_by){
	super::sstate.stats->nbr_bin_clses++;
	Literal imp 					= super::addBinAndSend(l0,l1);

	if(super::sstate.vinf.litFixTrue(imp)){ 	 return UNKNOWN_TrBool;}

	super::backjumpKeepLevel(0);		asrt(!super::openBCPImpls()); asrt(!super::sstate.vinf.litTrue(imp));

	return TrBool(super::sstate.vinf.litUkwn(imp) &&
					super::enqueue(imp,0) && (super::bcpLong()==super::BCP_SUCCESS));
}
*/
/*
LEADERTEMP
bool 	LEADERCLASS::deriveShorts(SharedClause& cls){
	// change
	return true;
	asrt(super::decisionLevel()==0);
	LOG_MEDIUM("Derive shorts of clause with size " << cls.size());
	Literal a_lits[cls.size()]; Literal b_lits[cls.size()];
	uint32_t asz = 0; uint32_t bsz = 0;

#ifdef MATRIX
	Comp ca,cb,comp;
	for(uint32_t i=0;i!=cls.size();++i){		asrt(!super::binDag.isLitRepl(cls[i]));
		LOG_DETAIL("Check lit" << cls[i].toDIMACS());
		if(!super::binDag.getComp(cls[i],comp))
				{LOG_MEDIUM("Invalid comp in deriveShorts"); return true;}
		if(ca.isInval() || ca==comp)	{ca=comp;a_lits[asz++]=cls[i];continue;}
		if(cb.isInval() || cb==comp)	{cb=comp;b_lits[bsz++]=cls[i];continue;}
		LOG_MEDIUM("Too many comps in deriveShorts"); return true;
	}
	asrt(cls.size()==asz+bsz);
	const bool one_comp = cb.isInval();
	if(!super::binDag.reachSame(ca,a_lits,asz)) {
		super::sstate.stats->nbr_comp_derive_cheap_fail++;
		LOG_MEDIUM("No sinks reached with comp " << ca.idx() << (one_comp?" single":" of two"));return true;}

	if(!one_comp && !super::binDag.reachSame(cb,b_lits,bsz)){
		super::sstate.stats->nbr_comp_derive_cheap_fail++;
		LOG_MEDIUM("No sinks reached with second comp " << cb.idx()); return true;}

#else // use lit roots
	Literal ra,rb; asrt(ra == rb && ra == INVALID_LIT);
	for(uint32_t i=0;i!=cls.size();++i){		asrt(!super::binDag.isLitRepl(cls[i]));
		LOG_DETAIL("Check lit" << cls[i].toDIMACS());
		Literal root = super::sstate.vinf.litRoot(cls[i]);
		if(ra==INVALID_LIT || ra==root)	{ra=root;a_lits[asz++]=cls[i];continue;}
		if(rb==INVALID_LIT || rb==root)	{rb=root;b_lits[bsz++]=cls[i];continue;}
		LOG_MEDIUM("Too many roots in deriveShorts"); return true;
	}

	asrt(cls.size()==asz+bsz);
	const bool one_comp = rb == INVALID_LIT;
#endif //MATRIX


	vector<Literal> common_a;
	super::collectCommonUkwn(a_lits,asz,common_a);
	if(common_a.empty()){
		super::sstate.stats->nbr_comp_derive_exp_fail++;
		LOG_MEDIUM("sink bits misleading"); return true;}

	if(one_comp){//only one component -> learn units
		LOG_BASIC("Hyper-Hyper created " << common_a.size() << " unit clauses");
		super::sstate.stats->nbr_comp_derive_unit+=common_a.size();
		return super::learnUnitClses(common_a);}

	vector<Literal> common_b;
	super::collectCommonUkwn(b_lits,bsz,common_b);
	if(common_b.empty()){
		super::sstate.stats->nbr_comp_derive_exp_fail++;
		LOG_MEDIUM("sink bits misleading"); return true;}

	//crossproduct of topmost common literals
	for(uint32_t i=common_a.size();i--;){
		Literal a = common_a[i];
		for(uint32_t j=common_b.size();j--;){
			super::sstate.stats->nbr_comp_derive_bin++;
			if(!importShort(a,common_b[j],super::sstate.s_info->user_id) ||
				(super::bcpLong()!=super::BCP_SUCCESS)) return false;
		}
	}
	LOG_BASIC("Hyper-Hyper created " << (common_a.size()*common_b.size()) << " binary clauses");
	return true;
}
*/


#endif /* CDCLCHIEF_HPP_ */
